from __future__ import print_function

import os
import string
from contextlib import contextmanager

import mock
import pytest
from click.testing import CliRunner

from dagster import (
    PartitionSetDefinition,
    PresetDefinition,
    ScheduleDefinition,
    Shape,
    check,
    lambda_solid,
    pipeline,
    repository,
    seven,
    solid,
)
from dagster.cli.run import run_list_command, run_wipe_command
from dagster.core.host_representation import ExternalPipeline
from dagster.core.instance import DagsterInstance
from dagster.core.launcher import RunLauncher
from dagster.core.storage.pipeline_run import PipelineRun
from dagster.core.test_utils import instance_for_test_tempdir
from dagster.core.types.loadable_target_origin import LoadableTargetOrigin
from dagster.grpc.server import GrpcServerProcess
from dagster.serdes import ConfigurableClass
from dagster.utils import file_relative_path, merge_dicts


def no_print(_):
    return None


@lambda_solid
def do_something():
    return 1


@lambda_solid
def do_input(x):
    return x


@pipeline(
    name='foo', preset_defs=[PresetDefinition(name="test", tags={'foo': 'bar'}),],
)
def foo_pipeline():
    do_input(do_something())


def define_foo_pipeline():
    return foo_pipeline


@pipeline(name='baz', description='Not much tbh')
def baz_pipeline():
    do_input()


def not_a_repo_or_pipeline_fn():
    return 'kdjfkjdf'


not_a_repo_or_pipeline = 123


def define_bar_schedules():
    return {
        'foo_schedule': ScheduleDefinition(
            "foo_schedule", cron_schedule="* * * * *", pipeline_name="test_pipeline", run_config={},
        ),
    }


def define_baz_partitions():
    return {
        'baz_partitions': PartitionSetDefinition(
            name='baz_partitions',
            pipeline_name='baz',
            partition_fn=lambda: string.digits,
            run_config_fn_for_partition=lambda partition: {
                'solids': {'do_input': {'inputs': {'x': {'value': partition.value}}}}
            },
        )
    }


@repository
def bar():
    return {
        'pipelines': {'foo': foo_pipeline, 'baz': baz_pipeline},
        'schedules': define_bar_schedules(),
        'partition_sets': define_baz_partitions(),
    }


@solid
def spew(context):
    context.log.info('HELLO WORLD')


@solid
def fail(context):
    raise Exception('I AM SUPPOSED TO FAIL')


@pipeline
def stdout_pipeline():
    spew()


@pipeline
def stderr_pipeline():
    fail()


class InMemoryRunLauncher(RunLauncher, ConfigurableClass):
    def __init__(self, inst_data=None):
        self._inst_data = inst_data
        self._queue = []

    def launch_run(self, instance, run, external_pipeline):
        check.inst_param(instance, 'instance', DagsterInstance)
        check.inst_param(run, 'run', PipelineRun)
        check.inst_param(external_pipeline, 'external_pipeline', ExternalPipeline)
        self._queue.append(run)
        return run

    def queue(self):
        return self._queue

    @classmethod
    def config_type(cls):
        return Shape({})

    @classmethod
    def from_config_value(cls, inst_data, config_value):
        return cls(inst_data=inst_data,)

    @property
    def inst_data(self):
        return self._inst_data

    def can_terminate(self, run_id):
        return False

    def terminate(self, run_id):
        check.not_implemented('Termintation not supported')


@contextmanager
def default_instance_tempdir(temp_dir, overrides=None):
    default_overrides = {
        'run_launcher': {
            'module': 'dagster_tests.cli_tests.command_tests.test_cli_commands',
            'class': 'InMemoryRunLauncher',
        }
    }
    with instance_for_test_tempdir(
        temp_dir, overrides=merge_dicts(default_overrides, (overrides if overrides else {}))
    ) as instance:
        with mock.patch('dagster.core.instance.DagsterInstance.get') as _instance:
            _instance.return_value = instance
            yield instance


@contextmanager
def default_instance(overrides=None):
    with seven.TemporaryDirectory() as temp_dir:
        with default_instance_tempdir(temp_dir, overrides) as instance:
            yield instance


def managed_grpc_instance():
    return default_instance(overrides={"opt_in": {"local_servers": True}})


@contextmanager
def args_with_instance(gen_instance, *args):
    with gen_instance as instance:
        yield args + (instance,)


def args_with_default_instance(*args):
    return args_with_instance(default_instance(), *args)


def args_with_managed_grpc_instance(*args):
    return args_with_instance(managed_grpc_instance(), *args)


@contextmanager
def grpc_server_bar_kwargs(pipeline_name=None):
    server_process = GrpcServerProcess(
        loadable_target_origin=LoadableTargetOrigin(
            python_file=file_relative_path(__file__, 'test_cli_commands.py'), attribute='bar'
        ),
    )
    with server_process.create_ephemeral_client() as client:
        args = {'grpc_host': client.host}
        if pipeline_name:
            args['pipeline'] = 'foo'
        if client.port:
            args['grpc_port'] = client.port
        if client.socket:
            args['grpc_socket'] = client.socket
        yield args
    server_process.wait()


@contextmanager
def python_bar_cli_args(pipeline_name=None):
    args = [
        '-m',
        'dagster_tests.cli_tests.command_tests.test_cli_commands',
        '-a',
        'bar',
    ]
    if pipeline_name:
        args.append('-p')
        args.append(pipeline_name)
    yield args


@contextmanager
def grpc_server_bar_cli_args(pipeline_name=None):
    server_process = GrpcServerProcess(
        loadable_target_origin=LoadableTargetOrigin(
            python_file=file_relative_path(__file__, 'test_cli_commands.py'), attribute='bar'
        ),
    )
    with server_process.create_ephemeral_client() as client:
        args = ['--grpc-host', client.host]
        if client.port:
            args.append('--grpc-port')
            args.append(client.port)
        if client.socket:
            args.append('--grpc-socket')
            args.append(client.socket)
        if pipeline_name:
            args.append('--pipeline')
            args.append(pipeline_name)

        yield args
    server_process.wait()


@contextmanager
def grpc_server_bar_pipeline_args():
    with default_instance() as instance:
        with grpc_server_bar_kwargs(pipeline_name='foo') as kwargs:
            yield kwargs, False, instance


# This iterates over a list of contextmanagers that can be used to contruct
# (cli_args, uses_legacy_repository_yaml_format, instance tuples)
def launch_command_contexts():
    for pipeline_target_args in valid_external_pipeline_target_args():
        yield args_with_default_instance(*pipeline_target_args)
        yield pytest.param(
            args_with_managed_grpc_instance(*pipeline_target_args), marks=pytest.mark.managed_grpc
        )
    yield pytest.param(grpc_server_bar_pipeline_args(), marks=pytest.mark.deployed_grpc)


def pipeline_python_origin_contexts():
    return [
        args_with_default_instance(pipeline_target_args)
        for pipeline_target_args in valid_pipeline_python_origin_target_args()
    ]


@contextmanager
def scheduler_instance(overrides=None):
    with seven.TemporaryDirectory() as temp_dir:
        with default_instance_tempdir(
            temp_dir,
            overrides=merge_dicts(
                {
                    'scheduler': {
                        'module': 'dagster.utils.test',
                        'class': 'FilesystemTestScheduler',
                        'config': {'base_dir': temp_dir},
                    }
                },
                overrides if overrides else {},
            ),
        ) as instance:
            yield instance


def managed_grpc_scheduler_instance():
    return scheduler_instance(overrides={"opt_in": {"local_servers": True}})


@contextmanager
def grpc_server_scheduler_cli_args():
    with scheduler_instance() as instance:
        with grpc_server_bar_cli_args() as args:
            yield args, instance


# Returns a list of contextmanagers that can be used to contruct
# (cli_args, instance) tuples for schedule calls
def schedule_command_contexts():
    return [
        args_with_instance(
            scheduler_instance(), ['-w', file_relative_path(__file__, 'workspace.yaml')]
        ),
        pytest.param(
            args_with_instance(
                managed_grpc_scheduler_instance(),
                ['-w', file_relative_path(__file__, 'workspace.yaml')],
            ),
            marks=pytest.mark.managed_grpc,
        ),
        pytest.param(grpc_server_scheduler_cli_args(), marks=pytest.mark.deployed_grpc),
    ]


# This iterates over a list of contextmanagers that can be used to contruct
# (cli_args, uses_legacy_repository_yaml_format, instance) tuples for backfill calls
def backfill_command_contexts():
    repo_args = {
        'noprompt': True,
        'workspace': (file_relative_path(__file__, 'repository_file.yaml'),),
    }
    return [
        args_with_instance(default_instance(), repo_args, True),
        pytest.param(
            args_with_instance(managed_grpc_instance(), repo_args, True),
            marks=pytest.mark.managed_grpc,
        ),
        pytest.param(grpc_server_backfill_args(), marks=pytest.mark.deployed_grpc),
    ]


@contextmanager
def grpc_server_backfill_args():
    with default_instance() as instance:
        with grpc_server_bar_kwargs() as args:
            yield merge_dicts(args, {'noprompt': True}), False, instance


def valid_pipeline_python_origin_target_args():
    return [
        {
            'workspace': None,
            'pipeline': 'foo',
            'python_file': file_relative_path(__file__, 'test_cli_commands.py'),
            'module_name': None,
            'attribute': 'bar',
        },
        {
            'workspace': None,
            'pipeline': 'foo',
            'python_file': file_relative_path(__file__, 'test_cli_commands.py'),
            'module_name': None,
            'attribute': 'bar',
            'working_directory': os.path.dirname(__file__),
        },
        {
            'workspace': None,
            'pipeline': 'foo',
            'python_file': None,
            'module_name': 'dagster_tests.cli_tests.command_tests.test_cli_commands',
            'attribute': 'bar',
        },
        {
            'workspace': None,
            'pipeline': None,
            'python_file': None,
            'module_name': 'dagster_tests.cli_tests.command_tests.test_cli_commands',
            'attribute': 'foo_pipeline',
        },
        {
            'workspace': None,
            'pipeline': None,
            'python_file': file_relative_path(__file__, 'test_cli_commands.py'),
            'module_name': None,
            'attribute': 'define_foo_pipeline',
        },
        {
            'workspace': None,
            'pipeline': None,
            'python_file': file_relative_path(__file__, 'test_cli_commands.py'),
            'module_name': None,
            'attribute': 'define_foo_pipeline',
            'working_directory': os.path.dirname(__file__),
        },
        {
            'workspace': None,
            'pipeline': None,
            'python_file': file_relative_path(__file__, 'test_cli_commands.py'),
            'module_name': None,
            'attribute': 'foo_pipeline',
        },
    ]


# [(cli_args, uses_legacy_repository_yaml_format)]
def valid_external_pipeline_target_args():
    return [
        (
            {
                'workspace': (file_relative_path(__file__, 'repository_file.yaml'),),
                'pipeline': 'foo',
                'python_file': None,
                'module_name': None,
                'attribute': None,
            },
            True,
        ),
        (
            {
                'workspace': (file_relative_path(__file__, 'repository_module.yaml'),),
                'pipeline': 'foo',
                'python_file': None,
                'module_name': None,
                'attribute': None,
            },
            True,
        ),
    ] + [(args, False) for args in valid_pipeline_python_origin_target_args()]


def valid_pipeline_python_origin_target_cli_args():
    return [
        ['-f', file_relative_path(__file__, 'test_cli_commands.py'), '-a', 'bar', '-p', 'foo'],
        [
            '-f',
            file_relative_path(__file__, 'test_cli_commands.py'),
            '-d',
            os.path.dirname(__file__),
            '-a',
            'bar',
            '-p',
            'foo',
        ],
        [
            '-m',
            'dagster_tests.cli_tests.command_tests.test_cli_commands',
            '-a',
            'bar',
            '-p',
            'foo',
        ],
        ['-m', 'dagster_tests.cli_tests.command_tests.test_cli_commands', '-a', 'foo_pipeline'],
        ['-f', file_relative_path(__file__, 'test_cli_commands.py'), '-a', 'define_foo_pipeline',],
        [
            '-f',
            file_relative_path(__file__, 'test_cli_commands.py'),
            '-d',
            os.path.dirname(__file__),
            '-a',
            'define_foo_pipeline',
        ],
    ]


# [(cli_args, uses_legacy_repository_yaml_format)]
def valid_external_pipeline_target_cli_args_no_preset():
    return [
        (['-w', file_relative_path(__file__, 'repository_file.yaml'), '-p', 'foo'], True),
        (['-w', file_relative_path(__file__, 'repository_module.yaml'), '-p', 'foo'], True),
        (['-w', file_relative_path(__file__, 'workspace.yaml'), '-p', 'foo'], False),
        (
            [
                '-w',
                file_relative_path(__file__, 'override.yaml'),
                '-w',
                file_relative_path(__file__, 'workspace.yaml'),
                '-p',
                'foo',
            ],
            False,
        ),
    ] + [(args, False) for args in valid_pipeline_python_origin_target_cli_args()]


def valid_external_pipeline_target_cli_args_with_preset():
    return valid_external_pipeline_target_cli_args_no_preset() + [
        (
            [
                '-f',
                file_relative_path(__file__, 'test_cli_commands.py'),
                '-d',
                os.path.dirname(__file__),
                '-a',
                'define_foo_pipeline',
                '--preset',
                'test',
            ],
            False,
        )
    ]


def test_run_list():
    runner = CliRunner()
    result = runner.invoke(run_list_command)
    assert result.exit_code == 0


def test_run_wipe_correct_delete_message():
    runner = CliRunner()
    result = runner.invoke(run_wipe_command, input="DELETE\n")
    assert 'Deleted all run history and event logs' in result.output
    assert result.exit_code == 0


def test_run_wipe_incorrect_delete_message():
    runner = CliRunner()
    result = runner.invoke(run_wipe_command, input="WRONG\n")
    assert 'Exiting without deleting all run history and event logs' in result.output
    assert result.exit_code == 0



import json

import pytest

from dagster import (
    Any,
    AssetKey,
    CompositeSolidDefinition,
    DependencyDefinition,
    InputDefinition,
    Int,
    OutputDefinition,
    PipelineDefinition,
    SolidInvocation,
    String,
    lambda_solid,
    solid,
)
from dagster.core.definitions import AssetMaterialization, Solid, create_run_config_schema
from dagster.core.definitions.dependency import SolidHandle, SolidOutputHandle
from dagster.core.errors import DagsterInvalidDefinitionError


def test_deps_equal():
    assert DependencyDefinition('foo') == DependencyDefinition('foo')
    assert DependencyDefinition('foo') != DependencyDefinition('bar')

    assert DependencyDefinition('foo', 'bar') == DependencyDefinition('foo', 'bar')
    assert DependencyDefinition('foo', 'bar') != DependencyDefinition('foo', 'quuz')


def test_solid_def():
    @lambda_solid
    def produce_string():
        return 'foo'

    @solid(
        input_defs=[InputDefinition('input_one', String)],
        output_defs=[OutputDefinition(Any)],
        config_schema={'another_field': Int},
    )
    def solid_one(_context, input_one):
        raise Exception('should not execute')

    pipeline_def = PipelineDefinition(
        solid_defs=[produce_string, solid_one],
        dependencies={'solid_one': {'input_one': DependencyDefinition('produce_string')}},
    )

    assert len(pipeline_def.solids[0].output_handles()) == 1

    assert isinstance(pipeline_def.solid_named('solid_one'), Solid)

    solid_one_solid = pipeline_def.solid_named('solid_one')

    assert solid_one_solid.has_input('input_one')

    assert isinstance(solid_one_solid.input_def_named('input_one'), InputDefinition)

    assert len(solid_one_solid.input_dict) == 1
    assert len(solid_one_solid.output_dict) == 1

    assert str(solid_one_solid.input_handle('input_one')) == (
        'SolidInputHandle(input_name="\'input_one\'", solid_name="\'solid_one\'")'
    )

    assert repr(solid_one_solid.input_handle('input_one')) == (
        'SolidInputHandle(input_name="\'input_one\'", solid_name="\'solid_one\'")'
    )

    assert str(solid_one_solid.output_handle('result')) == (
        'SolidOutputHandle(output_name="\'result\'", solid_name="\'solid_one\'")'
    )

    assert repr(solid_one_solid.output_handle('result')) == (
        'SolidOutputHandle(output_name="\'result\'", solid_name="\'solid_one\'")'
    )

    assert solid_one_solid.output_handle('result') == SolidOutputHandle(
        solid_one_solid, solid_one_solid.output_dict['result']
    )

    assert (
        len(pipeline_def.dependency_structure.input_to_upstream_outputs_for_solid('solid_one')) == 1
    )

    assert (
        len(
            pipeline_def.dependency_structure.output_to_downstream_inputs_for_solid(
                'produce_string'
            )
        )
        == 1
    )

    assert len(pipeline_def.dependency_structure.input_handles()) == 1

    assert len(pipeline_def.dependency_structure.items()) == 1


def test_solid_def_bad_input_name():
    with pytest.raises(DagsterInvalidDefinitionError, match='context is not allowed'):
        # pylint: disable=unused-variable
        @solid(input_defs=[InputDefinition('context', String)])
        def solid_one(_, _context):
            pass


def test_pipeline_types():
    @lambda_solid
    def produce_string():
        return 'foo'

    @solid(
        input_defs=[InputDefinition('input_one', String)],
        output_defs=[OutputDefinition(Any)],
        config_schema={'another_field': Int},
    )
    def solid_one(_context, input_one):
        raise Exception('should not execute')

    pipeline_def = PipelineDefinition(
        solid_defs=[produce_string, solid_one],
        dependencies={'solid_one': {'input_one': DependencyDefinition('produce_string')}},
    )

    run_config_schema = create_run_config_schema(pipeline_def)

    assert run_config_schema.has_config_type('String')
    assert run_config_schema.has_config_type('Int')
    assert not run_config_schema.has_config_type('SomeName')


def test_mapper_errors():
    @lambda_solid
    def solid_a():
        return 1

    with pytest.raises(DagsterInvalidDefinitionError) as excinfo_1:
        PipelineDefinition(
            solid_defs=[solid_a],
            dependencies={'solid_b': {'arg_a': DependencyDefinition('solid_a')}},
        )
    assert (
        str(excinfo_1.value)
        == 'Invalid dependencies: solid "solid_b" in dependency dictionary not found in solid list'
    )

    with pytest.raises(DagsterInvalidDefinitionError) as excinfo_2:
        PipelineDefinition(
            solid_defs=[solid_a],
            dependencies={
                SolidInvocation('solid_b', alias='solid_c'): {
                    'arg_a': DependencyDefinition('solid_a')
                }
            },
        )
    assert (
        str(excinfo_2.value)
        == 'Invalid dependencies: solid "solid_b" (aliased by "solid_c" in dependency dictionary) not found in solid list'
    )


def test_materialization():
    assert isinstance(AssetMaterialization('foo', 'foo.txt'), AssetMaterialization)


def test_materialization_assign_label_from_asset_key():
    mat = AssetMaterialization(asset_key=AssetKey(['foo', 'bar']))
    assert mat.label == 'foo.bar'


def test_rehydrate_solid_handle():
    h = SolidHandle.from_dict({'name': 'foo', 'parent': None})
    assert h.name == 'foo'
    assert h.parent is None

    h = SolidHandle.from_dict(json.loads(json.dumps(h._asdict())))
    assert h.name == 'foo'
    assert h.parent is None

    h = SolidHandle.from_dict({'name': 'foo', 'parent': ['bar', None]})
    assert h.name == 'foo'
    assert isinstance(h.parent, SolidHandle)
    assert h.parent.name == 'bar'
    assert h.parent.parent is None

    h = SolidHandle.from_dict(json.loads(json.dumps(h._asdict())))
    assert h.name == 'foo'
    assert isinstance(h.parent, SolidHandle)
    assert h.parent.name == 'bar'
    assert h.parent.parent is None

    h = SolidHandle.from_dict({'name': 'foo', 'parent': ['bar', ['baz', None]]})
    assert h.name == 'foo'
    assert isinstance(h.parent, SolidHandle)
    assert h.parent.name == 'bar'
    assert isinstance(h.parent.parent, SolidHandle)
    assert h.parent.parent.name == 'baz'
    assert h.parent.parent.parent is None

    h = SolidHandle.from_dict(json.loads(json.dumps(h._asdict())))
    assert h.name == 'foo'
    assert isinstance(h.parent, SolidHandle)
    assert h.parent.name == 'bar'
    assert isinstance(h.parent.parent, SolidHandle)
    assert h.parent.parent.name == 'baz'
    assert h.parent.parent.parent is None


def test_cycle_detect():
    @lambda_solid
    def return_one():
        return 1

    @lambda_solid
    def add(a, b):
        return a + b

    with pytest.raises(DagsterInvalidDefinitionError, match="Circular dependencies exist"):
        PipelineDefinition(
            solid_defs=[return_one, add],
            dependencies={
                SolidInvocation('add', alias='first'): {
                    'a': DependencyDefinition('return_one'),
                    'b': DependencyDefinition('second'),
                },
                SolidInvocation('add', alias='second'): {
                    'a': DependencyDefinition('first'),
                    'b': DependencyDefinition('return_one'),
                },
            },
        )

    with pytest.raises(DagsterInvalidDefinitionError, match="Circular dependencies exist"):
        CompositeSolidDefinition(
            name='circletron',
            solid_defs=[return_one, add],
            dependencies={
                SolidInvocation('add', alias='first'): {
                    'a': DependencyDefinition('return_one'),
                    'b': DependencyDefinition('second'),
                },
                SolidInvocation('add', alias='second'): {
                    'a': DependencyDefinition('first'),
                    'b': DependencyDefinition('return_one'),
                },
            },
        )



from collections import namedtuple

from dagster.config import Field, Selector
from dagster.config.config_type import ALL_CONFIG_BUILTINS, Array, ConfigType
from dagster.config.field import check_opt_field_param
from dagster.config.field_utils import Shape
from dagster.config.iterate_types import iterate_config_types
from dagster.core.errors import DagsterInvalidDefinitionError
from dagster.core.types.dagster_type import ALL_RUNTIME_BUILTINS, construct_dagster_type_dictionary
from dagster.utils import check, ensure_single_item

from .dependency import DependencyStructure, Solid, SolidHandle, SolidInputHandle
from .logger import LoggerDefinition
from .mode import ModeDefinition
from .resource import ResourceDefinition
from .solid import CompositeSolidDefinition, ISolidDefinition, SolidDefinition


def _is_selector_field_optional(config_type):
    check.inst_param(config_type, 'config_type', ConfigType)
    if len(config_type.fields) > 1:
        return False
    else:
        _name, field = ensure_single_item(config_type.fields)
        return not field.is_required


def define_resource_dictionary_cls(resource_defs):
    check.dict_param(resource_defs, 'resource_defs', key_type=str, value_type=ResourceDefinition)

    fields = {}
    for resource_name, resource_def in resource_defs.items():
        if resource_def.config_schema:
            fields[resource_name] = Field(Shape({'config': resource_def.config_schema}))

    return Shape(fields=fields)


def remove_none_entries(ddict):
    return {k: v for k, v in ddict.items() if v is not None}


def define_solid_config_cls(config_schema, inputs_field, outputs_field):
    check_opt_field_param(config_schema, 'config_schema')
    check_opt_field_param(inputs_field, 'inputs_field')
    check_opt_field_param(outputs_field, 'outputs_field')

    return Shape(
        remove_none_entries(
            {'config': config_schema, 'inputs': inputs_field, 'outputs': outputs_field}
        ),
    )


class EnvironmentClassCreationData(
    namedtuple(
        'EnvironmentClassCreationData',
        'pipeline_name solids dependency_structure mode_definition logger_defs',
    )
):
    def __new__(cls, pipeline_name, solids, dependency_structure, mode_definition, logger_defs):
        return super(EnvironmentClassCreationData, cls).__new__(
            cls,
            pipeline_name=check.str_param(pipeline_name, 'pipeline_name'),
            solids=check.list_param(solids, 'solids', of_type=Solid),
            dependency_structure=check.inst_param(
                dependency_structure, 'dependency_structure', DependencyStructure
            ),
            mode_definition=check.inst_param(mode_definition, 'mode_definition', ModeDefinition),
            logger_defs=check.dict_param(
                logger_defs, 'logger_defs', key_type=str, value_type=LoggerDefinition
            ),
        )


def define_logger_dictionary_cls(creation_data):
    check.inst_param(creation_data, 'creation_data', EnvironmentClassCreationData)

    fields = {}

    for logger_name, logger_definition in creation_data.logger_defs.items():
        fields[logger_name] = Field(
            Shape(remove_none_entries({'config': logger_definition.config_schema}),),
            is_required=False,
        )

    return Shape(fields)


def define_environment_cls(creation_data):
    check.inst_param(creation_data, 'creation_data', EnvironmentClassCreationData)

    return Shape(
        fields=remove_none_entries(
            {
                'solids': Field(
                    define_solid_dictionary_cls(
                        creation_data.solids, creation_data.dependency_structure,
                    )
                ),
                'storage': Field(
                    define_storage_config_cls(creation_data.mode_definition), is_required=False,
                ),
                'intermediate_storage': Field(
                    define_intermediate_storage_config_cls(creation_data.mode_definition),
                    is_required=False,
                ),
                'execution': Field(
                    define_executor_config_cls(creation_data.mode_definition), is_required=False,
                ),
                'loggers': Field(define_logger_dictionary_cls(creation_data)),
                'resources': Field(
                    define_resource_dictionary_cls(creation_data.mode_definition.resource_defs)
                ),
            }
        ),
    )


def define_storage_config_cls(mode_definition):
    check.inst_param(mode_definition, 'mode_definition', ModeDefinition)

    fields = {}

    for storage_def in mode_definition.system_storage_defs:
        fields[storage_def.name] = Field(
            Shape(
                fields={'config': storage_def.config_schema} if storage_def.config_schema else {},
            )
        )

    return Selector(fields)


def define_intermediate_storage_config_cls(mode_definition):
    check.inst_param(mode_definition, 'mode_definition', ModeDefinition)

    fields = {}

    for intermediate_storage_def in mode_definition.intermediate_storage_defs:
        fields[intermediate_storage_def.name] = Field(
            Shape(
                fields={'config': intermediate_storage_def.config_schema}
                if intermediate_storage_def.config_schema
                else {},
            )
        )

    return Selector(fields)


def define_executor_config_cls(mode_definition):
    check.inst_param(mode_definition, 'mode_definition', ModeDefinition)

    fields = {}

    for executor_def in mode_definition.executor_defs:
        fields[executor_def.name] = Field(
            Shape(
                fields={'config': executor_def.config_schema} if executor_def.config_schema else {},
            )
        )

    return Selector(fields)


def get_inputs_field(solid, handle, dependency_structure):
    check.inst_param(solid, 'solid', Solid)
    check.inst_param(handle, 'handle', SolidHandle)
    check.inst_param(dependency_structure, 'dependency_structure', DependencyStructure)

    if not solid.definition.has_configurable_inputs:
        return None

    inputs_field_fields = {}
    for name, inp in solid.definition.input_dict.items():
        if inp.dagster_type.loader:
            inp_handle = SolidInputHandle(solid, inp)
            # If this input is not satisfied by a dependency you must
            # provide it via config
            if not dependency_structure.has_deps(inp_handle) and not solid.container_maps_input(
                name
            ):

                inputs_field_fields[name] = Field(
                    inp.dagster_type.loader.schema_type,
                    is_required=(not solid.definition.input_has_default(name)),
                )

    if not inputs_field_fields:
        return None

    return Field(Shape(inputs_field_fields))


def get_outputs_field(solid, handle):
    check.inst_param(solid, 'solid', Solid)
    check.inst_param(handle, 'handle', SolidHandle)

    solid_def = solid.definition

    if not solid_def.has_configurable_outputs:
        return None

    output_dict_fields = {}
    for name, out in solid_def.output_dict.items():
        if out.dagster_type.materializer:
            output_dict_fields[name] = Field(
                out.dagster_type.materializer.schema_type, is_required=False
            )

    output_entry_dict = Shape(output_dict_fields)

    return Field(Array(output_entry_dict), is_required=False)


def filtered_system_dict(fields):
    return Field(Shape(remove_none_entries(fields)))


def construct_leaf_solid_config(solid, handle, dependency_structure, config_schema):
    check.inst_param(solid, 'solid', Solid)
    check.inst_param(handle, 'handle', SolidHandle)
    check.inst_param(dependency_structure, 'dependency_structure', DependencyStructure)

    return filtered_system_dict(
        {
            'inputs': get_inputs_field(solid, handle, dependency_structure),
            'outputs': get_outputs_field(solid, handle),
            'config': config_schema,
        }
    )


def define_isolid_field(solid, handle, dependency_structure):
    check.inst_param(solid, 'solid', Solid)
    check.inst_param(handle, 'handle', SolidHandle)

    # All solids regardless of compositing status get the same inputs and outputs
    # config. The only thing the varies is on extra element of configuration
    # 1) Vanilla solid definition: a 'config' key with the config_schema as the value
    # 2) Composite with field mapping: a 'config' key with the config_schema of
    #    the config mapping (via CompositeSolidDefinition#config_schema)
    # 3) Composite without field mapping: a 'solids' key with recursively defined
    #    solids dictionary
    # 4) `configured` composite with field mapping: a 'config' key with the config_schema that was
    #    provided when `configured` was called (via CompositeSolidDefinition#config_schema)

    if isinstance(solid.definition, SolidDefinition):
        return construct_leaf_solid_config(
            solid, handle, dependency_structure, solid.definition.config_schema
        )

    composite_def = check.inst(solid.definition, CompositeSolidDefinition)

    if composite_def.has_config_mapping:
        # has_config_mapping covers cases 2 & 4 from above (only config mapped composite solids can
        # be `configured`)...
        return construct_leaf_solid_config(
            solid,
            handle,
            dependency_structure,
            # ...and in both cases, the correct schema for 'config' key is exposed by this property:
            composite_def.config_schema,
        )
        # This case omits a 'solids' key, thus if a composite solid is `configured` or has a field
        # mapping, the user cannot stub any config, inputs, or outputs for inner (child) solids.
    else:
        return filtered_system_dict(
            {
                'inputs': get_inputs_field(solid, handle, dependency_structure),
                'outputs': get_outputs_field(solid, handle),
                'solids': Field(
                    define_solid_dictionary_cls(
                        composite_def.solids, composite_def.dependency_structure, handle,
                    )
                ),
            }
        )


def define_solid_dictionary_cls(solids, dependency_structure, parent_handle=None):
    check.list_param(solids, 'solids', of_type=Solid)
    check.inst_param(dependency_structure, 'dependency_structure', DependencyStructure)
    check.opt_inst_param(parent_handle, 'parent_handle', SolidHandle)

    fields = {}
    for solid in solids:
        if solid.definition.has_config_entry:
            fields[solid.name] = define_isolid_field(
                solid, SolidHandle(solid.name, parent_handle), dependency_structure,
            )

    return Shape(fields)


def iterate_solid_def_config_types(solid_def):

    if isinstance(solid_def, SolidDefinition):
        if solid_def.config_schema:
            for config_type in iterate_config_types(solid_def.config_schema.config_type):
                yield config_type
    elif isinstance(solid_def, CompositeSolidDefinition):
        for solid in solid_def.solids:
            for config_type in iterate_solid_def_config_types(solid.definition):
                yield config_type

    else:
        check.invariant('Unexpected ISolidDefinition type {type}'.format(type=type(solid_def)))


def _gather_all_schemas(solid_defs):
    dagster_types = construct_dagster_type_dictionary(solid_defs)
    for dagster_type in list(dagster_types.values()) + list(ALL_RUNTIME_BUILTINS):
        if dagster_type.loader:
            for ct in iterate_config_types(dagster_type.loader.schema_type):
                yield ct
        if dagster_type.materializer:
            for ct in iterate_config_types(dagster_type.materializer.schema_type):
                yield ct


def _gather_all_config_types(solid_defs, environment_type):
    check.list_param(solid_defs, 'solid_defs', ISolidDefinition)
    check.inst_param(environment_type, 'environment_type', ConfigType)

    for solid_def in solid_defs:
        for config_type in iterate_solid_def_config_types(solid_def):
            yield config_type

    for config_type in iterate_config_types(environment_type):
        yield config_type


def construct_config_type_dictionary(solid_defs, environment_type):
    check.list_param(solid_defs, 'solid_defs', ISolidDefinition)
    check.inst_param(environment_type, 'environment_type', ConfigType)

    type_dict_by_name = {t.given_name: t for t in ALL_CONFIG_BUILTINS if t.given_name}
    type_dict_by_key = {t.key: t for t in ALL_CONFIG_BUILTINS}
    all_types = list(_gather_all_config_types(solid_defs, environment_type)) + list(
        _gather_all_schemas(solid_defs)
    )

    for config_type in all_types:
        name = config_type.given_name
        if name and name in type_dict_by_name:
            if type(config_type) is not type(type_dict_by_name[name]):
                raise DagsterInvalidDefinitionError(
                    (
                        'Type names must be unique. You have constructed two different '
                        'instances of types with the same name "{name}".'
                    ).format(name=name)
                )
        elif name:
            type_dict_by_name[name] = config_type

        type_dict_by_key[config_type.key] = config_type

    return type_dict_by_name, type_dict_by_key



from abc import ABCMeta, abstractmethod

import six

from dagster import check
from dagster.core.errors import DagsterInvalidSubsetError
from dagster.core.selector import parse_solid_selection


class ExecutablePipeline(six.with_metaclass(ABCMeta)):
    @abstractmethod
    def get_definition(self):
        pass

    @abstractmethod
    def subset_for_execution(self, solid_selection):
        pass


class InMemoryExecutablePipeline(ExecutablePipeline, object):
    def __init__(self, pipeline_def, solid_selection=None, solids_to_execute=None):
        self._pipeline_def = pipeline_def
        self._solid_selection = solid_selection
        self._solids_to_execute = solids_to_execute

    def get_definition(self):
        return self._pipeline_def

    def _resolve_solid_selection(self, solid_selection):
        # resolve a list of solid selection queries to a frozenset of qualified solid names
        # e.g. ['foo_solid+'] to {'foo_solid', 'bar_solid'}
        check.list_param(solid_selection, 'solid_selection', of_type=str)
        solids_to_execute = parse_solid_selection(self.get_definition(), solid_selection)
        if len(solids_to_execute) == 0:
            raise DagsterInvalidSubsetError(
                'No qualified solids to execute found for solid_selection={requested}'.format(
                    requested=solid_selection
                )
            )
        return solids_to_execute

    def _subset_for_execution(self, solids_to_execute, solid_selection=None):
        if self._pipeline_def.is_subset_pipeline:
            return InMemoryExecutablePipeline(
                self._pipeline_def.parent_pipeline_def.get_pipeline_subset_def(solids_to_execute),
                solid_selection=solid_selection,
                solids_to_execute=solids_to_execute,
            )

        return InMemoryExecutablePipeline(
            self._pipeline_def.get_pipeline_subset_def(solids_to_execute),
            solid_selection=solid_selection,
            solids_to_execute=solids_to_execute,
        )

    def subset_for_execution(self, solid_selection):
        # take a list of solid queries and resolve the queries to names of solids to execute
        check.list_param(solid_selection, 'solid_selection', of_type=str)

        solids_to_execute = self._resolve_solid_selection(solid_selection)
        return self._subset_for_execution(solids_to_execute, solid_selection)

    def subset_for_execution_from_existing_pipeline(self, solids_to_execute):
        # take a frozenset of resolved solid names from an existing pipeline run
        # so there's no need to parse the selection
        check.set_param(solids_to_execute, 'solids_to_execute', of_type=str)

        return self._subset_for_execution(solids_to_execute)

    @property
    def solid_selection(self):
        # a list of solid queries provided by the user
        return self._solid_selection  # List[str]

    @property
    def solids_to_execute(self):
        # a frozenset which contains the names of the solids to execute
        return self._solids_to_execute  # FrozenSet[str]



from collections import namedtuple
from enum import Enum

from dagster import check
from dagster.utils.error import SerializableErrorInfo

from .config_type import ConfigTypeKind
from .snap import ConfigFieldSnap, ConfigTypeSnap
from .stack import EvaluationStack, get_friendly_path_info, get_friendly_path_msg
from .traversal_context import ContextData
from .type_printer import print_config_type_key_to_string


class DagsterEvaluationErrorReason(Enum):
    RUNTIME_TYPE_MISMATCH = 'RUNTIME_TYPE_MISMATCH'
    MISSING_REQUIRED_FIELD = 'MISSING_REQUIRED_FIELD'
    MISSING_REQUIRED_FIELDS = 'MISSING_REQUIRED_FIELDS'
    FIELD_NOT_DEFINED = 'FIELD_NOT_DEFINED'
    FIELDS_NOT_DEFINED = 'FIELDS_NOT_DEFINED'
    SELECTOR_FIELD_ERROR = 'SELECTOR_FIELD_ERROR'
    FAILED_POST_PROCESSING = 'FAILED_POST_PROCESSING'


class FieldsNotDefinedErrorData(namedtuple('_FieldsNotDefinedErrorData', 'field_names')):
    def __new__(cls, field_names):
        return super(FieldsNotDefinedErrorData, cls).__new__(
            cls, check.list_param(field_names, 'field_names', of_type=str)
        )


class FieldNotDefinedErrorData(namedtuple('_FieldNotDefinedErrorData', 'field_name')):
    def __new__(cls, field_name):
        return super(FieldNotDefinedErrorData, cls).__new__(
            cls, check.str_param(field_name, 'field_name')
        )


class MissingFieldErrorData(namedtuple('_MissingFieldErrorData', 'field_name field_snap')):
    def __new__(cls, field_name, field_snap):
        return super(MissingFieldErrorData, cls).__new__(
            cls,
            check.str_param(field_name, 'field_name'),
            check.inst_param(field_snap, 'field_snap', ConfigFieldSnap),
        )


class MissingFieldsErrorData(namedtuple('_MissingFieldErrorData', 'field_names field_snaps')):
    def __new__(cls, field_names, field_snaps):
        return super(MissingFieldsErrorData, cls).__new__(
            cls,
            check.list_param(field_names, 'field_names', of_type=str),
            check.list_param(field_snaps, 'field_snaps', of_type=ConfigFieldSnap),
        )


class RuntimeMismatchErrorData(
    namedtuple('_RuntimeMismatchErrorData', 'config_type_snap value_rep')
):
    def __new__(cls, config_type_snap, value_rep):
        check.inst_param(config_type_snap, 'config_type', ConfigTypeSnap)
        return super(RuntimeMismatchErrorData, cls).__new__(
            cls, config_type_snap, check.str_param(value_rep, 'value_rep'),
        )


class SelectorTypeErrorData(
    namedtuple('_SelectorTypeErrorData', 'config_type_snap incoming_fields')
):
    def __new__(cls, config_type_snap, incoming_fields):
        check.inst_param(config_type_snap, 'config_type_snap', ConfigTypeSnap)
        check.param_invariant(config_type_snap.kind == ConfigTypeKind.SELECTOR, 'config_type')
        return super(SelectorTypeErrorData, cls).__new__(
            cls,
            config_type_snap,
            check.list_param(incoming_fields, 'incoming_fields', of_type=str),
        )


ERROR_DATA_TYPES = (
    FieldNotDefinedErrorData,
    FieldsNotDefinedErrorData,
    MissingFieldErrorData,
    MissingFieldsErrorData,
    RuntimeMismatchErrorData,
    SelectorTypeErrorData,
    SerializableErrorInfo,
)


class EvaluationError(namedtuple('_EvaluationError', 'stack reason message error_data')):
    def __new__(cls, stack, reason, message, error_data):
        return super(EvaluationError, cls).__new__(
            cls,
            check.inst_param(stack, 'stack', EvaluationStack),
            check.inst_param(reason, 'reason', DagsterEvaluationErrorReason),
            check.str_param(message, 'message'),
            check.inst_param(error_data, 'error_data', ERROR_DATA_TYPES),
        )


def _get_type_msg(type_in_context):
    if type_in_context.given_name is None:
        return ''
    else:
        return ' on type "{type_name}"'.format(type_name=type_in_context.given_name)


def create_dict_type_mismatch_error(context, config_value):
    check.inst_param(context, 'context', ContextData)

    path_msg, _path = get_friendly_path_info(context.stack)

    return EvaluationError(
        stack=context.stack,
        reason=DagsterEvaluationErrorReason.RUNTIME_TYPE_MISMATCH,
        message='Value {path_msg} must be dict. Expected: "{type_name}".'.format(
            path_msg=path_msg,
            type_name=print_config_type_key_to_string(
                context.config_schema_snapshot, context.config_type_key, with_lines=False
            ),
        ),
        error_data=RuntimeMismatchErrorData(
            config_type_snap=context.config_type_snap, value_rep=repr(config_value)
        ),
    )


def create_fields_not_defined_error(context, undefined_fields):
    check.inst_param(context, 'context', ContextData)
    check_config_type_in_context_has_fields(context, 'context')
    check.list_param(undefined_fields, 'undefined_fields', of_type=str)

    available_fields = sorted(context.config_type_snap.field_names)
    undefined_fields = sorted(undefined_fields)

    return EvaluationError(
        stack=context.stack,
        reason=DagsterEvaluationErrorReason.FIELDS_NOT_DEFINED,
        message=(
            'Fields "{undefined_fields}" are not defined {path_msg}. Available '
            'fields: "{available_fields}."'
        ).format(
            undefined_fields=undefined_fields,
            path_msg=get_friendly_path_msg(context.stack),
            available_fields=available_fields,
        ),
        error_data=FieldsNotDefinedErrorData(field_names=undefined_fields),
    )


def create_enum_type_mismatch_error(context, config_value):
    check.inst_param(context, 'context', ContextData)

    return EvaluationError(
        stack=context.stack,
        reason=DagsterEvaluationErrorReason.RUNTIME_TYPE_MISMATCH,
        message='Value {path_msg} for enum type {type_name} must be a string'.format(
            type_name=context.config_type_snap.given_name,
            path_msg=get_friendly_path_msg(context.stack),
        ),
        error_data=RuntimeMismatchErrorData(context.config_type_snap, repr(config_value)),
    )


def create_enum_value_missing_error(context, config_value):
    check.inst_param(context, 'context', ContextData)

    return EvaluationError(
        stack=context.stack,
        reason=DagsterEvaluationErrorReason.RUNTIME_TYPE_MISMATCH,
        message='Value {path_msg} not in enum type {type_name} got {config_value}'.format(
            config_value=config_value,
            type_name=context.config_type_snap.given_name,
            path_msg=get_friendly_path_msg(context.stack),
        ),
        error_data=RuntimeMismatchErrorData(context.config_type_snap, repr(config_value)),
    )


def check_config_type_in_context_has_fields(context, param_name):
    check.param_invariant(ConfigTypeKind.has_fields(context.config_type_snap.kind), param_name)


def create_field_not_defined_error(context, received_field):
    check.inst_param(context, 'context', ContextData)
    check_config_type_in_context_has_fields(context, 'context')
    check.str_param(received_field, 'received_field')

    return EvaluationError(
        stack=context.stack,
        reason=DagsterEvaluationErrorReason.FIELD_NOT_DEFINED,
        message='Undefined field "{received}" {path_msg}. Expected: "{type_name}".'.format(
            path_msg=get_friendly_path_msg(context.stack),
            type_name=print_config_type_key_to_string(
                context.config_schema_snapshot, context.config_type_key, with_lines=False
            ),
            received=received_field,
        ),
        error_data=FieldNotDefinedErrorData(field_name=received_field),
    )


def create_array_error(context, config_value):
    check.inst_param(context, 'context', ContextData)
    check.param_invariant(context.config_type_snap.kind == ConfigTypeKind.ARRAY, 'config_type')

    return EvaluationError(
        stack=context.stack,
        reason=DagsterEvaluationErrorReason.RUNTIME_TYPE_MISMATCH,
        message='Value {path_msg} must be list. Expected: {type_name}'.format(
            path_msg=get_friendly_path_msg(context.stack),
            type_name=print_config_type_key_to_string(
                context.config_schema_snapshot, context.config_type_key, with_lines=False
            ),
        ),
        error_data=RuntimeMismatchErrorData(context.config_type_snap, repr(config_value)),
    )


def create_missing_required_field_error(context, expected_field):
    check.inst_param(context, 'context', ContextData)
    check_config_type_in_context_has_fields(context, 'context')

    return EvaluationError(
        stack=context.stack,
        reason=DagsterEvaluationErrorReason.MISSING_REQUIRED_FIELD,
        message=(
            'Missing required field "{expected}" {path_msg}. Available Fields: '
            '"{available_fields}".'
        ).format(
            expected=expected_field,
            path_msg=get_friendly_path_msg(context.stack),
            available_fields=sorted(context.config_type_snap.field_names),
        ),
        error_data=MissingFieldErrorData(
            field_name=expected_field, field_snap=context.config_type_snap.get_field(expected_field)
        ),
    )


def create_missing_required_fields_error(context, missing_fields):
    check.inst_param(context, 'context', ContextData)
    check_config_type_in_context_has_fields(context, 'context')

    missing_fields = sorted(missing_fields)
    missing_field_snaps = list(map(context.config_type_snap.get_field, missing_fields))

    return EvaluationError(
        stack=context.stack,
        reason=DagsterEvaluationErrorReason.MISSING_REQUIRED_FIELDS,
        message='Missing required fields "{missing_fields}" {path_msg}".'.format(
            missing_fields=missing_fields, path_msg=get_friendly_path_msg(context.stack)
        ),
        error_data=MissingFieldsErrorData(
            field_names=missing_fields, field_snaps=missing_field_snaps
        ),
    )


def create_scalar_error(context, config_value):
    check.inst_param(context, 'context', ContextData)

    return EvaluationError(
        stack=context.stack,
        reason=DagsterEvaluationErrorReason.RUNTIME_TYPE_MISMATCH,
        message='Invalid scalar {path_msg}. Value "{config_value}" of type '
        '"{type}" is not valid for expected type "{type_name}".'.format(
            path_msg=get_friendly_path_msg(context.stack),
            type_name=context.config_type_snap.given_name,
            config_value=config_value,
            type=type(config_value),
        ),
        error_data=RuntimeMismatchErrorData(context.config_type_snap, repr(config_value)),
    )


def create_selector_multiple_fields_error(context, config_value):
    check.inst_param(context, 'context', ContextData)

    defined_fields = sorted(context.config_type_snap.field_names)
    incoming_fields = sorted(list(config_value.keys()))

    return EvaluationError(
        stack=context.stack,
        reason=DagsterEvaluationErrorReason.SELECTOR_FIELD_ERROR,
        message=(
            'You can only specify a single field {path_msg}. You specified {incoming_fields}. '
            'The available fields are {defined_fields}'
        ).format(
            incoming_fields=incoming_fields,
            defined_fields=defined_fields,
            path_msg=get_friendly_path_msg(context.stack),
        ),
        error_data=SelectorTypeErrorData(
            config_type_snap=context.config_type_snap, incoming_fields=incoming_fields
        ),
    )


def create_selector_multiple_fields_no_field_selected_error(context):
    check.inst_param(context, 'context', ContextData)

    defined_fields = sorted(context.config_type_snap.field_names)

    return EvaluationError(
        stack=context.stack,
        reason=DagsterEvaluationErrorReason.SELECTOR_FIELD_ERROR,
        message=(
            'Must specify a field {path_msg} if more than one field is defined. '
            'Defined fields: {defined_fields}'
        ).format(defined_fields=defined_fields, path_msg=get_friendly_path_msg(context.stack)),
        error_data=SelectorTypeErrorData(
            config_type_snap=context.config_type_snap, incoming_fields=[]
        ),
    )


def create_selector_type_error(context, config_value):
    check.inst_param(context, 'context', ContextData)

    return EvaluationError(
        stack=context.stack,
        reason=DagsterEvaluationErrorReason.RUNTIME_TYPE_MISMATCH,
        message='Value for selector type {path_msg} must be a dict'.format(
            path_msg=get_friendly_path_msg(context.stack)
        ),
        error_data=RuntimeMismatchErrorData(
            config_type_snap=context.config_type_snap, value_rep=repr(config_value)
        ),
    )


def create_selector_unspecified_value_error(context):
    check.inst_param(context, 'context', ContextData)

    defined_fields = sorted(list(context.config_type.fields.keys()))

    return EvaluationError(
        stack=context.stack,
        reason=DagsterEvaluationErrorReason.SELECTOR_FIELD_ERROR,
        message=(
            'Must specify the required field {path_msg}. Defined fields: {defined_fields}'
        ).format(defined_fields=defined_fields, path_msg=get_friendly_path_msg(context.stack)),
        error_data=SelectorTypeErrorData(
            config_type_snap=context.config_type_snap, incoming_fields=[]
        ),
    )


def create_none_not_allowed_error(context):
    check.inst_param(context, 'context', ContextData)

    return EvaluationError(
        stack=context.stack,
        reason=DagsterEvaluationErrorReason.RUNTIME_TYPE_MISMATCH,
        message='Value {path_msg} must not be None.'.format(
            path_msg=get_friendly_path_msg(context.stack),
        ),
        error_data=RuntimeMismatchErrorData(context.config_type_snap, repr(None)),
    )


def create_failed_post_processing_error(context, original_value, error_data):
    check.inst_param(context, 'context', ContextData)
    check.inst_param(error_data, 'error_data', SerializableErrorInfo)

    return EvaluationError(
        stack=context.stack,
        reason=DagsterEvaluationErrorReason.FAILED_POST_PROCESSING,
        message='Post processing {path_msg} of original value {original_value} failed:\n{error}'.format(
            path_msg=get_friendly_path_msg(context.stack),
            original_value=original_value,
            error=error_data.to_string(),
        ),
        error_data=error_data,
    )


class PostProcessingError(Exception):
    '''
    This is part of the formal API for implementing post_process
    methods on config types. Throw this error to indicate a
    that post processing cannot happen, and that the user
    must make a configuration and environment change in
    order resolve.
    '''



from collections import namedtuple

from dagster import check
from dagster.core.errors import DagsterInvalidDefinitionError
from dagster.core.instance import DagsterInstance
from dagster.core.storage.pipeline_run import PipelineRun
from dagster.core.storage.tags import check_tags
from dagster.utils import merge_dicts

from .mode import DEFAULT_MODE_NAME


class ScheduleExecutionContext(namedtuple('ScheduleExecutionContext', 'instance')):
    '''Schedule-specific execution context.

    An instance of this class is made available as the first argument to various ScheduleDefinition
    functions. It is passed as the first argument to ``run_config_fn``, ``tags_fn``,
    and ``should_execute``.

    Attributes:
        instance (DagsterInstance): The instance configured to run the schedule
    '''

    def __new__(
        cls, instance,
    ):

        return super(ScheduleExecutionContext, cls).__new__(
            cls, check.inst_param(instance, 'instance', DagsterInstance),
        )


class ScheduleDefinition(object):
    '''Define a schedule that targets a pipeline

    Args:
        name (str): The name of the schedule to create.
        cron_schedule (str): A valid cron string specifying when the schedule will run, e.g.,
            '45 23 * * 6' for a schedule that runs at 11:45 PM every Saturday.
        pipeline_name (str): The name of the pipeline to execute when the schedule runs.
        run_config (Optional[Dict]): The environment config that parameterizes this execution,
            as a dict.
        run_config_fn (Callable[[ScheduleExecutionContext], [Dict]]): A function that takes a
            ScheduleExecutionContext object and returns the environment configuration that
            parameterizes this execution, as a dict. You may set only one of ``run_config``
            and ``run_config_fn``.
        tags (Optional[Dict[str, str]]): A dictionary of tags (string key-value pairs) to attach
            to the scheduled runs.
        tags_fn (Optional[Callable[[ScheduleExecutionContext], Optional[Dict[str, str]]]]): A
            function that generates tags to attach to the schedules runs. Takes a
            :py:class:`~dagster.ScheduleExecutionContext` and returns a dictionary of tags (string
            key-value pairs). You may set only one of ``tags`` and ``tags_fn``.
        solid_selection (Optional[List[str]]): A list of solid subselection (including single
            solid names) to execute when the schedule runs. e.g. ``['*some_solid+', 'other_solid']``
        mode (Optional[str]): The mode to apply when executing this schedule. (default: 'default')
        should_execute (Optional[Callable[[ScheduleExecutionContext], bool]]): A function that runs
            at schedule execution time to determine whether a schedule should execute or skip. Takes
            a :py:class:`~dagster.ScheduleExecutionContext` and returns a boolean (``True`` if the
            schedule should execute). Defaults to a function that always returns ``True``.
        environment_vars (Optional[dict[str, str]]): The environment variables to set for the
            schedule
    '''

    __slots__ = [
        '_name',
        '_cron_schedule',
        '_execution_params',
        '_run_config_fn',
        '_run_config',
        '_tags',
        '_mode',
        '_pipeline_name',
        '_solid_selection',
        '_tags_fn',
        '_should_execute',
        '_environment_vars',
    ]

    def __init__(
        self,
        name,
        cron_schedule,
        pipeline_name,
        run_config=None,
        run_config_fn=None,
        tags=None,
        tags_fn=None,
        solid_selection=None,
        mode="default",
        should_execute=None,
        environment_vars=None,
    ):

        self._name = check.str_param(name, 'name')
        self._cron_schedule = check.str_param(cron_schedule, 'cron_schedule')
        self._pipeline_name = check.str_param(pipeline_name, 'pipeline_name')
        self._run_config = check.opt_dict_param(run_config, 'run_config')
        self._tags = check.opt_dict_param(tags, 'tags', key_type=str, value_type=str)

        check.opt_callable_param(tags_fn, 'tags_fn')
        self._solid_selection = check.opt_nullable_list_param(
            solid_selection, 'solid_selection', of_type=str
        )
        self._mode = check.opt_str_param(mode, 'mode', DEFAULT_MODE_NAME)
        check.opt_callable_param(should_execute, 'should_execute')
        self._environment_vars = check.opt_dict_param(
            environment_vars, 'environment_vars', key_type=str, value_type=str
        )

        if run_config_fn and run_config:
            raise DagsterInvalidDefinitionError(
                'Attempted to provide both run_config_fn and run_config as arguments'
                ' to ScheduleDefinition. Must provide only one of the two.'
            )

        if not run_config and not run_config_fn:
            run_config_fn = lambda _context: {}
        self._run_config_fn = run_config_fn

        if tags_fn and tags:
            raise DagsterInvalidDefinitionError(
                'Attempted to provide both tags_fn and tags as arguments'
                ' to ScheduleDefinition. Must provide only one of the two.'
            )

        if not tags and not tags_fn:
            tags_fn = lambda _context: {}
        self._tags_fn = tags_fn

        if not should_execute:
            should_execute = lambda _context: True
        self._should_execute = should_execute

    @property
    def name(self):
        return self._name

    @property
    def cron_schedule(self):
        return self._cron_schedule

    @property
    def environment_vars(self):
        return self._environment_vars

    @property
    def mode(self):
        return self._mode

    def get_run_config(self, context):
        check.inst_param(context, 'context', ScheduleExecutionContext)
        if self._run_config:
            return self._run_config
        return self._run_config_fn(context)

    def get_tags(self, context):
        check.inst_param(context, 'context', ScheduleExecutionContext)
        if self._tags:
            tags = self._tags
            check_tags(tags, 'tags')
        else:
            tags = self._tags_fn(context)
            # These tags are checked in _tags_fn_wrapper

        tags = merge_dicts(tags, PipelineRun.tags_for_schedule(self))

        return tags

    def should_execute(self, context):
        check.inst_param(context, 'context', ScheduleExecutionContext)
        return self._should_execute(context)

    @property
    def pipeline_name(self):
        return self._pipeline_name

    @property
    def solid_selection(self):
        return self._solid_selection



import re

import pytest

from dagster import (
    DagsterInstance,
    DagsterInvariantViolationError,
    DagsterStepOutputNotFoundError,
    InputDefinition,
    Output,
    OutputDefinition,
    execute_pipeline,
    pipeline,
    reconstructable,
    solid,
)


def test_multiple_outputs():
    @solid(
        name='multiple_outputs',
        input_defs=[],
        output_defs=[OutputDefinition(name='output_one'), OutputDefinition(name='output_two')],
    )
    def multiple_outputs(_):
        yield Output(output_name='output_one', value='foo')
        yield Output(output_name='output_two', value='bar')

    @pipeline
    def multiple_outputs_pipeline():
        multiple_outputs()

    result = execute_pipeline(multiple_outputs_pipeline)
    solid_result = result.solid_result_list[0]

    assert solid_result.solid.name == 'multiple_outputs'
    assert solid_result.output_value('output_one') == 'foo'
    assert solid_result.output_value('output_two') == 'bar'

    with pytest.raises(
        DagsterInvariantViolationError,
        match="Output 'not_defined' not defined in solid 'multiple_outputs'",
    ):
        solid_result.output_value('not_defined')


def test_wrong_multiple_output():
    @solid(
        name='multiple_outputs', input_defs=[], output_defs=[OutputDefinition(name='output_one')]
    )
    def multiple_outputs(_):
        yield Output(output_name='mismatch', value='foo')

    @pipeline
    def wrong_multiple_outputs_pipeline():
        multiple_outputs()

    with pytest.raises(DagsterInvariantViolationError):
        execute_pipeline(wrong_multiple_outputs_pipeline)


def test_multiple_outputs_of_same_name_disallowed():
    # make this illegal until it is supported

    @solid(
        name='multiple_outputs', input_defs=[], output_defs=[OutputDefinition(name='output_one')]
    )
    def multiple_outputs(_):
        yield Output(output_name='output_one', value='foo')
        yield Output(output_name='output_one', value='foo')

    @pipeline
    def muptiple_outputs_pipeline():
        multiple_outputs()

    with pytest.raises(DagsterInvariantViolationError):
        execute_pipeline(muptiple_outputs_pipeline)


def define_multi_out():
    @solid(
        name='multiple_outputs',
        input_defs=[],
        output_defs=[
            OutputDefinition(name='output_one'),
            OutputDefinition(name='output_two', is_required=False),
        ],
    )
    def multiple_outputs(_):
        yield Output(output_name='output_one', value='foo')

    @solid(name='downstream_one', input_defs=[InputDefinition('some_input')], output_defs=[])
    def downstream_one(_, some_input):
        del some_input

    @solid
    def downstream_two(_, some_input):
        del some_input
        raise Exception('do not call me')

    @pipeline
    def multiple_outputs_only_emit_one_pipeline():
        output_one, output_two = multiple_outputs()
        downstream_one(output_one)
        downstream_two(output_two)

    return multiple_outputs_only_emit_one_pipeline


def test_multiple_outputs_only_emit_one():
    result = execute_pipeline(define_multi_out())
    assert result.success

    solid_result = result.result_for_solid('multiple_outputs')
    assert set(solid_result.output_values.keys()) == set(['output_one'])

    with pytest.raises(
        DagsterInvariantViolationError,
        match="Output 'not_defined' not defined in solid 'multiple_outputs'",
    ):
        solid_result.output_value('not_defined')

    with pytest.raises(DagsterInvariantViolationError, match='Did not find result output_two'):
        solid_result.output_value('output_two')

    with pytest.raises(
        DagsterInvariantViolationError,
        match=re.escape(
            'Tried to get result for solid \'not_present\' in '
            '\'multiple_outputs_only_emit_one_pipeline\'. No such top level solid.'
        ),
    ):
        result.result_for_solid('not_present')

    assert result.result_for_solid('downstream_two').skipped


def test_multiple_outputs_only_emit_one_multiproc():
    pipe = reconstructable(define_multi_out)
    result = execute_pipeline(
        pipe,
        run_config={'storage': {'filesystem': {}}, 'execution': {'multiprocess': {}}},
        instance=DagsterInstance.local_temp(),
    )
    assert result.success

    solid_result = result.result_for_solid('multiple_outputs')
    assert set(solid_result.output_values.keys()) == set(['output_one'])

    with pytest.raises(
        DagsterInvariantViolationError,
        match="Output 'not_defined' not defined in solid 'multiple_outputs'",
    ):
        solid_result.output_value('not_defined')

    with pytest.raises(DagsterInvariantViolationError, match='Did not find result output_two'):
        solid_result.output_value('output_two')

    with pytest.raises(
        DagsterInvariantViolationError,
        match=re.escape(
            'Tried to get result for solid \'not_present\' in '
            '\'multiple_outputs_only_emit_one_pipeline\'. No such top level solid.'
        ),
    ):
        result.result_for_solid('not_present')

    assert result.result_for_solid('downstream_two').skipped


def test_missing_non_optional_output_fails():
    @solid(
        name='multiple_outputs',
        input_defs=[],
        output_defs=[OutputDefinition(name='output_one'), OutputDefinition(name='output_two')],
    )
    def multiple_outputs(_):
        yield Output(output_name='output_one', value='foo')

    @pipeline
    def missing_non_optional_pipeline():
        multiple_outputs()

    with pytest.raises(DagsterStepOutputNotFoundError):
        execute_pipeline(missing_non_optional_pipeline)



from dagster import check
from dagster.core.host_representation.external_data import (
    ExternalScheduleExecutionData,
    ExternalScheduleExecutionErrorData,
)
from dagster.core.host_representation.handle import RepositoryHandle
from dagster.core.types.loadable_target_origin import LoadableTargetOrigin
from dagster.grpc.types import ExternalScheduleExecutionArgs, ScheduleExecutionDataMode

from .utils import execute_unary_api_cli_command


def sync_get_external_schedule_execution_data(
    instance, repository_handle, schedule_name, schedule_execution_data_mode
):

    check.inst_param(repository_handle, 'repository_handle', RepositoryHandle)
    check.str_param(schedule_name, 'schedule_name')
    check.inst_param(
        schedule_execution_data_mode, 'schedule_execution_data_mode', ScheduleExecutionDataMode
    )

    origin = repository_handle.get_origin()

    return check.inst(
        execute_unary_api_cli_command(
            origin.executable_path,
            'schedule_config',
            ExternalScheduleExecutionArgs(
                repository_origin=origin,
                instance_ref=instance.get_ref(),
                schedule_name=schedule_name,
                schedule_execution_data_mode=schedule_execution_data_mode,
            ),
        ),
        (ExternalScheduleExecutionData, ExternalScheduleExecutionErrorData),
    )


def sync_get_external_schedule_execution_data_ephemeral_grpc(
    instance, repository_handle, schedule_name, schedule_execution_data_mode
):
    from dagster.grpc.client import ephemeral_grpc_api_client

    origin = repository_handle.get_origin()
    with ephemeral_grpc_api_client(
        LoadableTargetOrigin(executable_path=origin.executable_path)
    ) as api_client:
        return sync_get_external_schedule_execution_data_grpc(
            api_client, instance, repository_handle, schedule_name, schedule_execution_data_mode,
        )


def sync_get_external_schedule_execution_data_grpc(
    api_client, instance, repository_handle, schedule_name, schedule_execution_data_mode,
):
    check.inst_param(repository_handle, 'repository_handle', RepositoryHandle)
    check.str_param(schedule_name, 'schedule_name')
    check.inst_param(
        schedule_execution_data_mode, 'schedule_execution_data_mode', ScheduleExecutionDataMode
    )

    origin = repository_handle.get_origin()

    return check.inst(
        api_client.external_schedule_execution(
            external_schedule_execution_args=ExternalScheduleExecutionArgs(
                repository_origin=origin,
                instance_ref=instance.get_ref(),
                schedule_name=schedule_name,
                schedule_execution_data_mode=schedule_execution_data_mode,
            )
        ),
        (ExternalScheduleExecutionData, ExternalScheduleExecutionErrorData),
    )



from contextlib import contextmanager

import pytest


@contextmanager
def assert_no_warnings():
    # https://stackoverflow.com/questions/45671803/how-to-use-pytest-to-assert-no-warning-is-raised
    with pytest.warns(None) as record:
        yield
    assert len(record) == 0, 'Unexpected warnings: {warnings}'.format(
        warnings=[str(record[i]) for i in range(len(record))]
    )



import os

from dagster import check
from dagster.serdes import ConfigurableClass, ConfigurableClassData


class LocalArtifactStorage(ConfigurableClass):
    def __init__(self, base_dir, inst_data=None):
        self._base_dir = base_dir
        self._inst_data = check.opt_inst_param(inst_data, 'inst_data', ConfigurableClassData)

    @property
    def inst_data(self):
        return self._inst_data

    @property
    def base_dir(self):
        return self._base_dir

    def file_manager_dir(self, run_id):
        check.str_param(run_id, 'run_id')
        return os.path.join(self.base_dir, 'storage', run_id, 'files')

    def intermediates_dir(self, run_id):
        return os.path.join(self.base_dir, 'storage', run_id, '')

    @property
    def schedules_dir(self):
        return os.path.join(self.base_dir, 'schedules')

    @staticmethod
    def from_config_value(inst_data, config_value):
        return LocalArtifactStorage(inst_data=inst_data, **config_value)

    @classmethod
    def config_type(cls):
        return {'base_dir': str}



import pytest

from dagster.core.definitions.utils import (
    check_valid_name,
    config_from_files,
    config_from_pkg_resources,
    config_from_yaml_strings,
)
from dagster.core.errors import DagsterInvalidDefinitionError, DagsterInvariantViolationError
from dagster.utils import file_relative_path


def test_check_valid_name():
    assert check_valid_name('a') == 'a'

    with pytest.raises(DagsterInvalidDefinitionError):
        assert check_valid_name('has a space')

    with pytest.raises(DagsterInvalidDefinitionError):
        assert check_valid_name('')

    with pytest.raises(DagsterInvalidDefinitionError):
        assert check_valid_name('context')


def test_config_from_files():
    run_config = config_from_files(
        config_files=[file_relative_path(__file__, './definitions_tests/pass_env.yaml')]
    )

    assert run_config == {'solids': {'can_fail': {'config': {'error': False}}}}

    with pytest.raises(DagsterInvariantViolationError):
        config_from_files(config_files=[file_relative_path(__file__, 'not_a_file.yaml')])

    with pytest.raises(DagsterInvariantViolationError):
        config_from_files(
            config_files=[
                file_relative_path(__file__, './definitions_tests/test_repository_definition.py')
            ]
        )


def test_config_from_yaml_strings():
    a = '''
foo:
  bar: 1
baz: 3
'''
    b = '''
foo:
  one: "one"
other: 4
'''
    c = '''
final: "result"
'''

    run_config = config_from_yaml_strings([a, b, c])
    assert run_config == {
        'foo': {'bar': 1, 'one': 'one'},
        'baz': 3,
        'other': 4,
        'final': 'result',
    }

    with pytest.raises(
        DagsterInvariantViolationError, match='Encountered error attempting to parse yaml'
    ):
        config_from_yaml_strings(['--- `'])

    run_config = config_from_yaml_strings([])
    assert run_config == {}


def test_config_from_pkg_resources():
    good = (
        'dagster_tests.core_tests.definitions_tests',
        'pass_env.yaml',
    )
    run_config = config_from_pkg_resources([good])
    assert run_config == {'solids': {'can_fail': {'config': {'error': False}}}}

    bad_defs = [
        ('dagster_tests.core_tests.definitions_tests', 'does_not_exist.yaml'),
        ('dagster_tests.core_tests.definitions_tests', 'bad_file_binary.yaml',),
    ]

    for bad_def in bad_defs:
        with pytest.raises(
            DagsterInvariantViolationError, match='Encountered error attempting to parse yaml',
        ):
            config_from_pkg_resources([bad_def])



import os
import sqlite3
from functools import update_wrapper

from dagster import check

from .sql import run_migrations_offline as run_migrations_offline_
from .sql import run_migrations_online as run_migrations_online_


def run_migrations_offline(*args, **kwargs):
    try:
        run_migrations_offline_(*args, **kwargs)
    except sqlite3.DatabaseError as exc:
        # This is to deal with concurrent execution -- if this table already exists thanks to a
        # race with another process, we are fine and can continue.
        if not 'table alembic_version already exists' in str(exc):
            raise


def run_migrations_online(*args, **kwargs):
    try:
        run_migrations_online_(*args, **kwargs)
    except (sqlite3.DatabaseError, sqlite3.OperationalError) as exc:
        # This is to deal with concurrent execution -- if this table already exists thanks to a
        # race with another process, we are fine and can continue.
        if not 'table alembic_version already exists' in str(exc):
            raise


update_wrapper(run_migrations_offline, run_migrations_offline_)

update_wrapper(run_migrations_online, run_migrations_online_)


def create_db_conn_string(base_dir, db_name):
    check.str_param(base_dir, 'base_dir')
    check.str_param(db_name, 'db_name')

    path_components = os.path.abspath(base_dir).split(os.sep)
    db_file = '{}.db'.format(db_name)
    return 'sqlite:///{}'.format('/'.join(path_components + [db_file]))



from abc import abstractmethod
from enum import Enum as PythonEnum
from functools import partial

import six

from dagster import check
from dagster.builtins import BuiltinEnum
from dagster.config.config_type import Array
from dagster.config.config_type import Noneable as ConfigNoneable
from dagster.core.definitions.events import TypeCheck
from dagster.core.errors import DagsterInvalidDefinitionError, DagsterInvariantViolationError
from dagster.core.storage.type_storage import TypeStoragePlugin
from dagster.serdes import whitelist_for_serdes
from dagster.utils.backcompat import canonicalize_backcompat_args, rename_warning

from .builtin_config_schemas import BuiltinSchemas
from .config_schema import DagsterTypeLoader, DagsterTypeMaterializer
from .marshal import PickleSerializationStrategy, SerializationStrategy


@whitelist_for_serdes
class DagsterTypeKind(PythonEnum):
    ANY = 'ANY'
    SCALAR = 'SCALAR'
    LIST = 'LIST'
    NOTHING = 'NOTHING'
    NULLABLE = 'NULLABLE'
    REGULAR = 'REGULAR'


class DagsterType(object):
    '''Define a type in dagster. These can be used in the inputs and outputs of solids.

    Args:
        type_check_fn (Callable[[TypeCheckContext, Any], [Union[bool, TypeCheck]]]):
            The function that defines the type check. It takes the value flowing
            through the input or output of the solid. If it passes, return either
            ``True`` or a :py:class:`~dagster.TypeCheck` with ``success`` set to ``True``. If it fails,
            return either ``False`` or a :py:class:`~dagster.TypeCheck` with ``success`` set to ``False``.
            The first argument must be named ``context`` (or, if unused, ``_``, ``_context``, or ``context_``).
            Use ``required_resource_keys`` for access to resources.
        key (Optional[str]): The unique key to identify types programatically.
            The key property always has a value. If you omit key to the argument
            to the init function, it instead receives the value of ``name``. If
            neither ``key`` nor ``name`` is provided, a ``CheckError`` is thrown.

            In the case of a generic type such as ``List`` or ``Optional``, this is
            generated programatically based on the type parameters.

            For most use cases, name should be set and the key argument should
            not be specified.
        name (Optional[str]): A unique name given by a user. If ``key`` is ``None``, ``key``
            becomes this value. Name is not given in a case where the user does
            not specify a unique name for this type, such as a generic class.
        description (Optional[str]): A markdown-formatted string, displayed in tooling.
        loader (Optional[DagsterTypeLoader]): An instance of a class that
            inherits from :py:class:`~dagster.DagsterTypeLoader` and can map config data to a value of
            this type. Specify this argument if you will need to shim values of this type using the
            config machinery. As a rule, you should use the
            :py:func:`@dagster_type_loader <dagster.dagster_type_loader>` decorator to construct
            these arguments.
        materializer (Optional[DagsterTypeMaterializer]): An instance of a class
            that inherits from :py:class:`~dagster.DagsterTypeMaterializer` and can persist values of
            this type. As a rule, you should use the
            :py:func:`@dagster_type_materializer <dagster.dagster_type_materializer>`
            decorator to construct these arguments.
        serialization_strategy (Optional[SerializationStrategy]): An instance of a class that
            inherits from :py:class:`~dagster.SerializationStrategy`. The default strategy for serializing
            this value when automatically persisting it between execution steps. You should set
            this value if the ordinary serialization machinery (e.g., pickle) will not be adequate
            for this type.
        auto_plugins (Optional[List[Type[TypeStoragePlugin]]]): If types must be serialized differently
            depending on the storage being used for intermediates, they should specify this
            argument. In these cases the serialization_strategy argument is not sufficient because
            serialization requires specialized API calls, e.g. to call an S3 API directly instead
            of using a generic file object. See ``dagster_pyspark.DataFrame`` for an example.
        required_resource_keys (Optional[Set[str]]): Resource keys required by the ``type_check_fn``.
        is_builtin (bool): Defaults to False. This is used by tools to display or
            filter built-in types (such as :py:class:`~dagster.String`, :py:class:`~dagster.Int`) to visually distinguish
            them from user-defined types. Meant for internal use.
        kind (DagsterTypeKind): Defaults to None. This is used to determine the kind of runtime type
            for InputDefinition and OutputDefinition type checking.
    '''

    def __init__(
        self,
        type_check_fn,
        key=None,
        name=None,
        is_builtin=False,
        description=None,
        loader=None,
        materializer=None,
        serialization_strategy=None,
        auto_plugins=None,
        required_resource_keys=None,
        kind=DagsterTypeKind.REGULAR,
        # Graveyard is below
        input_hydration_config=None,
        output_materialization_config=None,
    ):
        check.opt_str_param(key, 'key')
        check.opt_str_param(name, 'name')

        check.invariant(not (name is None and key is None), 'Must set key or name')

        if name is None:
            check.param_invariant(
                bool(key), 'key', 'If name is not provided, must provide key.',
            )
            self.key, self.name = key, None
        elif key is None:
            check.param_invariant(
                bool(name), 'name', 'If key is not provided, must provide name.',
            )
            self.key, self.name = name, name
        else:
            check.invariant(key and name)
            self.key, self.name = key, name

        self.description = check.opt_str_param(description, 'description')
        self.loader = canonicalize_backcompat_args(
            check.opt_inst_param(loader, 'loader', DagsterTypeLoader),
            'loader',
            check.opt_inst_param(
                input_hydration_config, 'input_hydration_config', DagsterTypeLoader
            ),
            'input_hydration_config',
            '0.10.0',
        )
        self.materializer = canonicalize_backcompat_args(
            check.opt_inst_param(materializer, 'materializer', DagsterTypeMaterializer),
            'materializer',
            check.opt_inst_param(
                output_materialization_config,
                'output_materialization_config',
                DagsterTypeMaterializer,
            ),
            'output_materialization_config',
            '0.10.0',
        )
        self.serialization_strategy = check.opt_inst_param(
            serialization_strategy,
            'serialization_strategy',
            SerializationStrategy,
            PickleSerializationStrategy(),
        )
        self.required_resource_keys = check.opt_set_param(
            required_resource_keys, 'required_resource_keys',
        )

        self._type_check_fn = check.callable_param(type_check_fn, 'type_check_fn')
        _validate_type_check_fn(self._type_check_fn, self.name)

        auto_plugins = check.opt_list_param(auto_plugins, 'auto_plugins', of_type=type)

        check.param_invariant(
            all(
                issubclass(auto_plugin_type, TypeStoragePlugin) for auto_plugin_type in auto_plugins
            ),
            'auto_plugins',
        )

        self.auto_plugins = auto_plugins

        self.is_builtin = check.bool_param(is_builtin, 'is_builtin')
        check.invariant(
            self.display_name is not None,
            'All types must have a valid display name, got None for key {}'.format(key),
        )

        self.kind = check.inst_param(kind, 'kind', DagsterTypeKind)

    def type_check(self, context, value):
        retval = self._type_check_fn(context, value)

        if not isinstance(retval, (bool, TypeCheck)):
            raise DagsterInvariantViolationError(
                (
                    'You have returned {retval} of type {retval_type} from the type '
                    'check function of type "{type_key}". Return value must be instance '
                    'of TypeCheck or a bool.'
                ).format(retval=repr(retval), retval_type=type(retval), type_key=self.key)
            )

        return TypeCheck(success=retval) if isinstance(retval, bool) else retval

    def __eq__(self, other):
        return isinstance(other, DagsterType) and self.key == other.key

    def __ne__(self, other):
        return not self.__eq__(other)

    @staticmethod
    def from_builtin_enum(builtin_enum):
        check.invariant(BuiltinEnum.contains(builtin_enum), 'must be member of BuiltinEnum')
        return _RUNTIME_MAP[builtin_enum]

    @property
    def display_name(self):
        return self.name

    @property
    def inner_types(self):
        return []

    @property
    def input_hydration_config(self):
        rename_warning('loader', 'input_hydration_config', '0.10.0')
        return self.loader

    @property
    def output_materialization_config(self):
        rename_warning('materializer', 'output_materialization_config', '0.10.0')
        return self.materializer

    @property
    def input_hydration_schema_key(self):
        rename_warning('loader_schema_key', 'input_hydration_schema_key', '0.10.0')
        return self.loader_schema_key

    @property
    def loader_schema_key(self):
        return self.loader.schema_type.key if self.loader else None

    @property
    def output_materialization_schema_key(self):
        rename_warning('materializer_schema_key', 'output_materialization_schema_key', '0.10.0')
        return self.materializer_schema_key

    @property
    def materializer_schema_key(self):
        return self.materializer.schema_type.key if self.materializer else None

    @property
    def type_param_keys(self):
        return []

    @property
    def is_nothing(self):
        return self.kind == DagsterTypeKind.NOTHING


def _validate_type_check_fn(fn, name):
    from dagster.seven import get_args

    args = get_args(fn)

    # py2 doesn't filter out self
    if len(args) >= 1 and args[0] == 'self':
        args = args[1:]

    if len(args) == 2:
        possible_names = {
            '_',
            'context',
            '_context',
            'context_',
        }
        if args[0] not in possible_names:
            DagsterInvalidDefinitionError(
                'type_check function on type "{name}" must have first '
                'argument named "context" (or _, _context, context_).'.format(name=name,)
            )
        return True

    raise DagsterInvalidDefinitionError(
        'type_check_fn argument on type "{name}" must take 2 arguments, '
        'received {count}.'.format(name=name, count=len(args))
    )


class BuiltinScalarDagsterType(DagsterType):
    def __init__(self, name, type_check_fn, *args, **kwargs):
        super(BuiltinScalarDagsterType, self).__init__(
            key=name,
            name=name,
            kind=DagsterTypeKind.SCALAR,
            type_check_fn=type_check_fn,
            is_builtin=True,
            *args,
            **kwargs
        )

    def type_check_fn(self, _context, value):
        return self.type_check_scalar_value(value)

    @abstractmethod
    def type_check_scalar_value(self, _value):
        raise NotImplementedError()


class _Int(BuiltinScalarDagsterType):
    def __init__(self):
        super(_Int, self).__init__(
            name='Int',
            loader=BuiltinSchemas.INT_INPUT,
            materializer=BuiltinSchemas.INT_OUTPUT,
            type_check_fn=self.type_check_fn,
        )

    def type_check_scalar_value(self, value):
        return _fail_if_not_of_type(value, six.integer_types, 'int')


def _typemismatch_error_str(value, expected_type_desc):
    return 'Value "{value}" of python type "{python_type}" must be a {type_desc}.'.format(
        value=value, python_type=type(value).__name__, type_desc=expected_type_desc
    )


def _fail_if_not_of_type(value, value_type, value_type_desc):

    if not isinstance(value, value_type):
        return TypeCheck(success=False, description=_typemismatch_error_str(value, value_type_desc))

    return TypeCheck(success=True)


class _String(BuiltinScalarDagsterType):
    def __init__(self):
        super(_String, self).__init__(
            name='String',
            loader=BuiltinSchemas.STRING_INPUT,
            materializer=BuiltinSchemas.STRING_OUTPUT,
            type_check_fn=self.type_check_fn,
        )

    def type_check_scalar_value(self, value):
        return _fail_if_not_of_type(value, six.string_types, 'string')


class _Float(BuiltinScalarDagsterType):
    def __init__(self):
        super(_Float, self).__init__(
            name='Float',
            loader=BuiltinSchemas.FLOAT_INPUT,
            materializer=BuiltinSchemas.FLOAT_OUTPUT,
            type_check_fn=self.type_check_fn,
        )

    def type_check_scalar_value(self, value):
        return _fail_if_not_of_type(value, float, 'float')


class _Bool(BuiltinScalarDagsterType):
    def __init__(self):
        super(_Bool, self).__init__(
            name='Bool',
            loader=BuiltinSchemas.BOOL_INPUT,
            materializer=BuiltinSchemas.BOOL_OUTPUT,
            type_check_fn=self.type_check_fn,
        )

    def type_check_scalar_value(self, value):
        return _fail_if_not_of_type(value, bool, 'bool')


class Anyish(DagsterType):
    def __init__(
        self,
        key,
        name,
        loader=None,
        materializer=None,
        serialization_strategy=None,
        is_builtin=False,
        description=None,
        auto_plugins=None,
    ):
        super(Anyish, self).__init__(
            key=key,
            name=name,
            kind=DagsterTypeKind.ANY,
            loader=loader,
            materializer=materializer,
            serialization_strategy=serialization_strategy,
            is_builtin=is_builtin,
            type_check_fn=self.type_check_method,
            description=description,
            auto_plugins=auto_plugins,
        )

    def type_check_method(self, _context, _value):
        return TypeCheck(success=True)


class _Any(Anyish):
    def __init__(self):
        super(_Any, self).__init__(
            key='Any',
            name='Any',
            loader=BuiltinSchemas.ANY_INPUT,
            materializer=BuiltinSchemas.ANY_OUTPUT,
            is_builtin=True,
        )


def create_any_type(
    name,
    loader=None,
    materializer=None,
    serialization_strategy=None,
    description=None,
    auto_plugins=None,
):
    return Anyish(
        key=name,
        name=name,
        description=description,
        loader=loader,
        materializer=materializer,
        serialization_strategy=serialization_strategy,
        auto_plugins=auto_plugins,
    )


class _Nothing(DagsterType):
    def __init__(self):
        super(_Nothing, self).__init__(
            key='Nothing',
            name='Nothing',
            kind=DagsterTypeKind.NOTHING,
            loader=None,
            materializer=None,
            type_check_fn=self.type_check_method,
            is_builtin=True,
        )

    def type_check_method(self, _context, value):
        if value is not None:
            return TypeCheck(
                success=False,
                description='Value must be None, got a {value_type}'.format(value_type=type(value)),
            )

        return TypeCheck(success=True)


class PythonObjectDagsterType(DagsterType):
    '''Define a type in dagster whose typecheck is an isinstance check.

    Specifically, the type can either be a single python type (e.g. int),
    or a tuple of types (e.g. (int, float)) which is treated as a union.

    Examples:
        .. code-block:: python
            ntype = PythonObjectDagsterType(python_type=int)
            assert ntype.name == 'int'
            assert_success(ntype, 1)
            assert_failure(ntype, 'a')
        ```

        .. code-block:: python
            ntype = PythonObjectDagsterType(python_type=(int, float))
            assert ntype.name == 'Union[int, float]'
            assert_success(ntype, 1)
            assert_success(ntype, 1.5)
            assert_failure(ntype, 'a')


    Args:
        python_type (Union[Type, Tuple[Type, ...]): The dagster typecheck function calls instanceof on
            this type.
        name (Optional[str]): Name the type. Defaults to the name of ``python_type``.
        key (Optional[str]): Key of the type. Defaults to name.
        description (Optional[str]): A markdown-formatted string, displayed in tooling.
        loader (Optional[DagsterTypeLoader]): An instance of a class that
            inherits from :py:class:`~dagster.DagsterTypeLoader` and can map config data to a value of
            this type. Specify this argument if you will need to shim values of this type using the
            config machinery. As a rule, you should use the
            :py:func:`@dagster_type_loader <dagster.dagster_type_loader>` decorator to construct
            these arguments.
        materializer (Optional[DagsterTypeMaterializer]): An instance of a class
            that inherits from :py:class:`~dagster.DagsterTypeMaterializer` and can persist values of
            this type. As a rule, you should use the
            :py:func:`@dagster_type_mate <dagster.dagster_type_mate>`
            decorator to construct these arguments.
        serialization_strategy (Optional[SerializationStrategy]): An instance of a class that
            inherits from :py:class:`SerializationStrategy`. The default strategy for serializing
            this value when automatically persisting it between execution steps. You should set
            this value if the ordinary serialization machinery (e.g., pickle) will not be adequate
            for this type.
        auto_plugins (Optional[List[Type[TypeStoragePlugin]]]): If types must be serialized differently
            depending on the storage being used for intermediates, they should specify this
            argument. In these cases the serialization_strategy argument is not sufficient because
            serialization requires specialized API calls, e.g. to call an S3 API directly instead
            of using a generic file object. See ``dagster_pyspark.DataFrame`` for an example.

    '''

    def __init__(self, python_type, key=None, name=None, **kwargs):
        if isinstance(python_type, tuple):
            self.python_type = check.tuple_param(
                python_type, 'python_type', of_type=tuple(check.type_types for item in python_type)
            )
            self.type_str = 'Union[{}]'.format(
                ', '.join(python_type.__name__ for python_type in python_type)
            )
        else:
            self.python_type = check.type_param(python_type, 'python_type')
            self.type_str = python_type.__name__
        name = check.opt_str_param(name, 'name', self.type_str)
        key = check.opt_str_param(key, 'key', name)
        super(PythonObjectDagsterType, self).__init__(
            key=key, name=name, type_check_fn=self.type_check_method, **kwargs
        )

    def type_check_method(self, _context, value):
        if not isinstance(value, self.python_type):
            return TypeCheck(
                success=False,
                description=(
                    'Value of type {value_type} failed type check for Dagster type {dagster_type}, '
                    'expected value to be of Python type {expected_type}.'
                ).format(
                    value_type=type(value), dagster_type=self.name, expected_type=self.type_str,
                ),
            )

        return TypeCheck(success=True)


class NoneableInputSchema(DagsterTypeLoader):
    def __init__(self, inner_dagster_type):
        self._inner_dagster_type = check.inst_param(
            inner_dagster_type, 'inner_dagster_type', DagsterType
        )
        check.param_invariant(inner_dagster_type.loader, 'inner_dagster_type')
        self._schema_type = ConfigNoneable(inner_dagster_type.loader.schema_type)

    @property
    def schema_type(self):
        return self._schema_type

    def construct_from_config_value(self, context, config_value):
        if config_value is None:
            return None
        return self._inner_dagster_type.loader.construct_from_config_value(context, config_value)


def _create_nullable_input_schema(inner_type):
    if not inner_type.loader:
        return None

    return NoneableInputSchema(inner_type)


class OptionalType(DagsterType):
    def __init__(self, inner_type):
        inner_type = resolve_dagster_type(inner_type)

        if inner_type is Nothing:
            raise DagsterInvalidDefinitionError(
                'Type Nothing can not be wrapped in List or Optional'
            )

        key = 'Optional.' + inner_type.key
        self.inner_type = inner_type
        super(OptionalType, self).__init__(
            key=key,
            name=None,
            kind=DagsterTypeKind.NULLABLE,
            type_check_fn=self.type_check_method,
            loader=_create_nullable_input_schema(inner_type),
        )

    @property
    def display_name(self):
        return self.inner_type.display_name + '?'

    def type_check_method(self, context, value):
        return (
            TypeCheck(success=True) if value is None else self.inner_type.type_check(context, value)
        )

    @property
    def inner_types(self):
        return [self.inner_type] + self.inner_type.inner_types

    @property
    def type_param_keys(self):
        return [self.inner_type.key]


class ListInputSchema(DagsterTypeLoader):
    def __init__(self, inner_dagster_type):
        self._inner_dagster_type = check.inst_param(
            inner_dagster_type, 'inner_dagster_type', DagsterType
        )
        check.param_invariant(inner_dagster_type.loader, 'inner_dagster_type')
        self._schema_type = Array(inner_dagster_type.loader.schema_type)

    @property
    def schema_type(self):
        return self._schema_type

    def construct_from_config_value(self, context, config_value):
        convert_item = partial(self._inner_dagster_type.loader.construct_from_config_value, context)
        return list(map(convert_item, config_value))


def _create_list_input_schema(inner_type):
    if not inner_type.loader:
        return None

    return ListInputSchema(inner_type)


class ListType(DagsterType):
    def __init__(self, inner_type):
        key = 'List.' + inner_type.key
        self.inner_type = inner_type
        super(ListType, self).__init__(
            key=key,
            name=None,
            kind=DagsterTypeKind.LIST,
            type_check_fn=self.type_check_method,
            loader=_create_list_input_schema(inner_type),
        )

    @property
    def display_name(self):
        return '[' + self.inner_type.display_name + ']'

    def type_check_method(self, context, value):
        value_check = _fail_if_not_of_type(value, list, 'list')
        if not value_check.success:
            return value_check

        for item in value:
            item_check = self.inner_type.type_check(context, item)
            if not item_check.success:
                return item_check

        return TypeCheck(success=True)

    @property
    def inner_types(self):
        return [self.inner_type] + self.inner_type.inner_types

    @property
    def type_param_keys(self):
        return [self.inner_type.key]


class DagsterListApi:
    def __getitem__(self, inner_type):
        check.not_none_param(inner_type, 'inner_type')
        return _List(resolve_dagster_type(inner_type))

    def __call__(self, inner_type):
        check.not_none_param(inner_type, 'inner_type')
        return _List(inner_type)


List = DagsterListApi()


def _List(inner_type):
    check.inst_param(inner_type, 'inner_type', DagsterType)
    if inner_type is Nothing:
        raise DagsterInvalidDefinitionError('Type Nothing can not be wrapped in List or Optional')
    return ListType(inner_type)


class Stringish(DagsterType):
    def __init__(self, key=None, name=None, **kwargs):
        name = check.opt_str_param(name, 'name', type(self).__name__)
        key = check.opt_str_param(key, 'key', name)
        super(Stringish, self).__init__(
            key=key,
            name=name,
            kind=DagsterTypeKind.SCALAR,
            type_check_fn=self.type_check_method,
            loader=BuiltinSchemas.STRING_INPUT,
            materializer=BuiltinSchemas.STRING_OUTPUT,
            **kwargs
        )

    def type_check_method(self, _context, value):
        return _fail_if_not_of_type(value, six.string_types, 'string')


def create_string_type(name, description=None):
    return Stringish(name=name, key=name, description=description)


Any = _Any()
Bool = _Bool()
Float = _Float()
Int = _Int()
String = _String()
Nothing = _Nothing()

_RUNTIME_MAP = {
    BuiltinEnum.ANY: Any,
    BuiltinEnum.BOOL: Bool,
    BuiltinEnum.FLOAT: Float,
    BuiltinEnum.INT: Int,
    BuiltinEnum.STRING: String,
    BuiltinEnum.NOTHING: Nothing,
}

_PYTHON_TYPE_TO_DAGSTER_TYPE_MAPPING_REGISTRY = {}
'''Python types corresponding to user-defined RunTime types created using @map_to_dagster_type or
as_dagster_type are registered here so that we can remap the Python types to runtime types.'''


def make_python_type_usable_as_dagster_type(python_type, dagster_type):
    '''
    Take any existing python type and map it to a dagster type (generally created with
    :py:class:`DagsterType <dagster.DagsterType>`) This can only be called once
    on a given python type.
    '''
    check.inst_param(dagster_type, 'dagster_type', DagsterType)
    if (
        _PYTHON_TYPE_TO_DAGSTER_TYPE_MAPPING_REGISTRY.get(python_type, dagster_type)
        is not dagster_type
    ):
        # This would be just a great place to insert a short URL pointing to the type system
        # documentation into the error message
        # https://github.com/dagster-io/dagster/issues/1831
        raise DagsterInvalidDefinitionError(
            (
                'A Dagster type has already been registered for the Python type '
                '{python_type}. make_python_type_usable_as_dagster_type can only '
                'be called once on a python type as it is registering a 1:1 mapping '
                'between that python type and a dagster type.'
            ).format(python_type=python_type)
        )

    _PYTHON_TYPE_TO_DAGSTER_TYPE_MAPPING_REGISTRY[python_type] = dagster_type


DAGSTER_INVALID_TYPE_ERROR_MESSAGE = (
    'Invalid type: dagster_type must be DagsterType, a python scalar, or a python type '
    'that has been marked usable as a dagster type via @usable_dagster_type or '
    'make_python_type_usable_as_dagster_type: got {dagster_type}{additional_msg}'
)


def resolve_dagster_type(dagster_type):
    # circular dep
    from .python_dict import PythonDict, Dict
    from .python_set import PythonSet, DagsterSetApi
    from .python_tuple import PythonTuple, DagsterTupleApi
    from .transform_typing import transform_typing_type
    from dagster.config.config_type import ConfigType
    from dagster.primitive_mapping import (
        remap_python_builtin_for_runtime,
        is_supported_runtime_python_builtin,
    )
    from dagster.utils.typing_api import is_typing_type

    check.invariant(
        not (isinstance(dagster_type, type) and issubclass(dagster_type, ConfigType)),
        'Cannot resolve a config type to a runtime type',
    )

    check.invariant(
        not (isinstance(dagster_type, type) and issubclass(dagster_type, DagsterType)),
        'Do not pass runtime type classes. Got {}'.format(dagster_type),
    )

    # First check to see if it part of python's typing library
    if is_typing_type(dagster_type):
        dagster_type = transform_typing_type(dagster_type)

    if isinstance(dagster_type, DagsterType):
        return dagster_type

    # Test for unhashable objects -- this is if, for instance, someone has passed us an instance of
    # a dict where they meant to pass dict or Dict, etc.
    try:
        hash(dagster_type)
    except TypeError:
        raise DagsterInvalidDefinitionError(
            DAGSTER_INVALID_TYPE_ERROR_MESSAGE.format(
                additional_msg=(
                    ', which isn\'t hashable. Did you pass an instance of a type instead of '
                    'the type?'
                ),
                dagster_type=str(dagster_type),
            )
        )

    if is_supported_runtime_python_builtin(dagster_type):
        return remap_python_builtin_for_runtime(dagster_type)

    if dagster_type is None:
        return Any

    if dagster_type in _PYTHON_TYPE_TO_DAGSTER_TYPE_MAPPING_REGISTRY:
        return _PYTHON_TYPE_TO_DAGSTER_TYPE_MAPPING_REGISTRY[dagster_type]

    if dagster_type is Dict:
        return PythonDict
    if isinstance(dagster_type, DagsterTupleApi):
        return PythonTuple
    if isinstance(dagster_type, DagsterSetApi):
        return PythonSet
    if isinstance(dagster_type, DagsterListApi):
        return List(Any)
    if BuiltinEnum.contains(dagster_type):
        return DagsterType.from_builtin_enum(dagster_type)
    if not isinstance(dagster_type, type):
        raise DagsterInvalidDefinitionError(
            DAGSTER_INVALID_TYPE_ERROR_MESSAGE.format(
                dagster_type=str(dagster_type), additional_msg='.'
            )
        )

    raise DagsterInvalidDefinitionError(
        '{dagster_type} is not a valid dagster type.'.format(dagster_type=dagster_type)
    )


ALL_RUNTIME_BUILTINS = list(_RUNTIME_MAP.values())


def construct_dagster_type_dictionary(solid_defs):
    type_dict_by_name = {t.name: t for t in ALL_RUNTIME_BUILTINS}
    type_dict_by_key = {t.key: t for t in ALL_RUNTIME_BUILTINS}
    for solid_def in solid_defs:
        for dagster_type in solid_def.all_dagster_types():
            # We don't do uniqueness check on key because with classes
            # like Array, Noneable, etc, those are ephemeral objects
            # and it is perfectly fine to have many of them.
            type_dict_by_key[dagster_type.key] = dagster_type

            if not dagster_type.name:
                continue

            if dagster_type.name not in type_dict_by_name:
                type_dict_by_name[dagster_type.name] = dagster_type
                continue

            if type_dict_by_name[dagster_type.name] is not dagster_type:
                raise DagsterInvalidDefinitionError(
                    (
                        'You have created two dagster types with the same name "{type_name}". '
                        'Dagster types have must have unique names.'
                    ).format(type_name=dagster_type.name)
                )

    return type_dict_by_key


class DagsterOptionalApi:
    def __getitem__(self, inner_type):
        check.not_none_param(inner_type, 'inner_type')
        return OptionalType(inner_type)


Optional = DagsterOptionalApi()



import os

import pytest

from dagster import Bool, Int, List, Optional, String, check
from dagster.core.definitions.events import ObjectStoreOperationType
from dagster.core.execution.plan.objects import StepOutputHandle
from dagster.core.instance import DagsterInstance
from dagster.core.storage.intermediate_storage import build_fs_intermediate_storage
from dagster.core.storage.type_storage import TypeStoragePlugin, TypeStoragePluginRegistry
from dagster.core.types.dagster_type import String as RuntimeString
from dagster.core.types.dagster_type import create_any_type, resolve_dagster_type
from dagster.core.types.marshal import SerializationStrategy
from dagster.core.utils import make_new_run_id
from dagster.utils import mkdir_p
from dagster.utils.test import yield_empty_pipeline_context


class UppercaseSerializationStrategy(SerializationStrategy):  # pylint: disable=no-init
    def serialize(self, value, write_file_obj):
        return write_file_obj.write(bytes(value.upper().encode('utf-8')))

    def deserialize(self, read_file_obj):
        return read_file_obj.read().decode('utf-8').lower()


class FancyStringFilesystemTypeStoragePlugin(TypeStoragePlugin):  # pylint:disable=no-init
    @classmethod
    def compatible_with_storage_def(cls, _):
        # Not needed for these tests
        raise NotImplementedError()

    @classmethod
    def set_intermediate_object(
        cls, intermediate_storage, context, dagster_type, step_output_handle, value
    ):
        paths = ['intermediates', step_output_handle.step_key, step_output_handle.output_name]
        paths.append(value)
        mkdir_p(os.path.join(intermediate_storage.root, *paths))

    @classmethod
    def get_intermediate_object(
        cls, intermediate_storage, context, dagster_type, step_output_handle
    ):
        paths = ['intermediates', step_output_handle.step_key, step_output_handle.output_name]
        return os.listdir(os.path.join(intermediate_storage.root, *paths))[0]


LowercaseString = create_any_type(
    'LowercaseString', serialization_strategy=UppercaseSerializationStrategy('uppercase')
)


def define_intermediate_storage(type_storage_plugin_registry=None):
    run_id = make_new_run_id()
    instance = DagsterInstance.ephemeral()
    intermediate_storage = build_fs_intermediate_storage(
        instance.intermediates_directory,
        run_id=run_id,
        type_storage_plugin_registry=type_storage_plugin_registry,
    )
    return run_id, instance, intermediate_storage


def test_file_system_intermediate_storage():
    _, _, intermediate_storage = define_intermediate_storage()

    assert (
        intermediate_storage.set_intermediate(
            None, Int, StepOutputHandle('return_one.compute'), 21
        ).op
        == ObjectStoreOperationType.SET_OBJECT
    )

    assert (
        intermediate_storage.rm_intermediate(None, StepOutputHandle('return_one.compute')).op
        == ObjectStoreOperationType.RM_OBJECT
    )

    assert (
        intermediate_storage.set_intermediate(
            None, Int, StepOutputHandle('return_one.compute'), 42
        ).op
        == ObjectStoreOperationType.SET_OBJECT
    )

    assert (
        intermediate_storage.get_intermediate(None, Int, StepOutputHandle('return_one.compute')).obj
        == 42
    )


def test_file_system_intermediate_storage_composite_types():
    _, _, intermediate_storage = define_intermediate_storage()

    assert intermediate_storage.set_intermediate(
        None, List[Bool], StepOutputHandle('return_true_lst.compute'), [True]
    )

    assert intermediate_storage.has_intermediate(None, StepOutputHandle('return_true_lst.compute'))

    assert intermediate_storage.get_intermediate(
        None, List[Bool], StepOutputHandle('return_true_lst.compute')
    ).obj == [True]


def test_file_system_intermediate_storage_with_custom_serializer():
    run_id, instance, intermediate_storage = define_intermediate_storage()
    with yield_empty_pipeline_context(run_id=run_id, instance=instance) as context:
        intermediate_storage.set_intermediate(
            context, LowercaseString, StepOutputHandle('a.b'), 'bar'
        )

        with open(
            os.path.join(intermediate_storage.root, 'intermediates', 'a.b', 'result'), 'rb'
        ) as fd:
            assert fd.read().decode('utf-8') == 'BAR'

        assert intermediate_storage.has_intermediate(context, StepOutputHandle('a.b'))
        assert (
            intermediate_storage.get_intermediate(
                context, LowercaseString, StepOutputHandle('a.b')
            ).obj
            == 'bar'
        )


def test_file_system_intermediate_storage_composite_types_with_custom_serializer_for_inner_type():
    run_id, instance, intermediate_storage = define_intermediate_storage()

    with yield_empty_pipeline_context(run_id=run_id, instance=instance) as context:

        intermediate_storage.set_intermediate(
            context, resolve_dagster_type(List[LowercaseString]), StepOutputHandle('baz'), ['list']
        )
        assert intermediate_storage.has_intermediate(context, StepOutputHandle('baz'))
        assert intermediate_storage.get_intermediate(
            context, resolve_dagster_type(List[Bool]), StepOutputHandle('baz')
        ).obj == ['list']


def test_file_system_intermediate_storage_with_type_storage_plugin():
    run_id, instance, intermediate_storage = define_intermediate_storage(
        type_storage_plugin_registry=TypeStoragePluginRegistry(
            [(RuntimeString, FancyStringFilesystemTypeStoragePlugin)]
        ),
    )

    with yield_empty_pipeline_context(run_id=run_id, instance=instance) as context:
        try:
            intermediate_storage.set_intermediate(
                context, RuntimeString, StepOutputHandle('obj_name'), 'hello'
            )

            assert intermediate_storage.has_intermediate(context, StepOutputHandle('obj_name'))
            assert (
                intermediate_storage.get_intermediate(
                    context, RuntimeString, StepOutputHandle('obj_name')
                )
                == 'hello'
            )

        finally:
            intermediate_storage.rm_intermediate(context, StepOutputHandle('obj_name'))


def test_file_system_intermediate_storage_with_composite_type_storage_plugin():
    run_id, _, intermediate_storage = define_intermediate_storage(
        type_storage_plugin_registry=TypeStoragePluginRegistry(
            [(RuntimeString, FancyStringFilesystemTypeStoragePlugin)]
        ),
    )

    with yield_empty_pipeline_context(run_id=run_id) as context:
        with pytest.raises(check.NotImplementedCheckError):
            intermediate_storage.set_intermediate(
                context, resolve_dagster_type(List[String]), StepOutputHandle('obj_name'), ['hello']
            )

    with yield_empty_pipeline_context(run_id=run_id) as context:
        with pytest.raises(check.NotImplementedCheckError):
            intermediate_storage.set_intermediate(
                context,
                resolve_dagster_type(Optional[String]),
                StepOutputHandle('obj_name'),
                ['hello'],
            )

    with yield_empty_pipeline_context(run_id=run_id) as context:
        with pytest.raises(check.NotImplementedCheckError):
            intermediate_storage.set_intermediate(
                context,
                resolve_dagster_type(List[Optional[String]]),
                StepOutputHandle('obj_name'),
                ['hello'],
            )

    with yield_empty_pipeline_context(run_id=run_id) as context:
        with pytest.raises(check.NotImplementedCheckError):
            intermediate_storage.set_intermediate(
                context,
                resolve_dagster_type(Optional[List[String]]),
                StepOutputHandle('obj_name'),
                ['hello'],
            )



from __future__ import print_function

import re

import click
import pytest
from click import UsageError
from click.testing import CliRunner

from dagster.cli.pipeline import execute_execute_command, pipeline_execute_command
from dagster.core.errors import DagsterInvariantViolationError
from dagster.core.test_utils import instance_for_test
from dagster.utils import file_relative_path, merge_dicts

from .test_cli_commands import (
    pipeline_python_origin_contexts,
    valid_pipeline_python_origin_target_cli_args,
)


def test_execute_mode_command():
    runner = CliRunner()

    with instance_for_test():
        add_result = runner_pipeline_execute(
            runner,
            [
                '-f',
                file_relative_path(__file__, '../../general_tests/test_repository.py'),
                '-a',
                'dagster_test_repository',
                '--config',
                file_relative_path(
                    __file__, '../../environments/multi_mode_with_resources/add_mode.yaml'
                ),
                '--mode',
                'add_mode',
                '-p',
                'multi_mode_with_resources',  # pipeline name
            ],
        )

        assert add_result

        mult_result = runner_pipeline_execute(
            runner,
            [
                '-f',
                file_relative_path(__file__, '../../general_tests/test_repository.py'),
                '-a',
                'dagster_test_repository',
                '--config',
                file_relative_path(
                    __file__, '../../environments/multi_mode_with_resources/mult_mode.yaml'
                ),
                '--mode',
                'mult_mode',
                '-p',
                'multi_mode_with_resources',  # pipeline name
            ],
        )

        assert mult_result

        double_adder_result = runner_pipeline_execute(
            runner,
            [
                '-f',
                file_relative_path(__file__, '../../general_tests/test_repository.py'),
                '-a',
                'dagster_test_repository',
                '--config',
                file_relative_path(
                    __file__, '../../environments/multi_mode_with_resources/double_adder_mode.yaml'
                ),
                '--mode',
                'double_adder_mode',
                '-p',
                'multi_mode_with_resources',  # pipeline name
            ],
        )

        assert double_adder_result


def test_execute_preset_command():
    with instance_for_test():
        runner = CliRunner()
        add_result = runner_pipeline_execute(
            runner,
            [
                '-f',
                file_relative_path(__file__, '../../general_tests/test_repository.py'),
                '-a',
                'dagster_test_repository',
                '--preset',
                'add',
                '-p',
                'multi_mode_with_resources',  # pipeline name
            ],
        )

        assert 'PIPELINE_SUCCESS' in add_result.output

        # Can't use --preset with --config
        bad_res = runner.invoke(
            pipeline_execute_command,
            [
                '-f',
                file_relative_path(__file__, '../../general_tests/test_repository.py'),
                '-a',
                'dagster_test_repository',
                '--preset',
                'add',
                '--config',
                file_relative_path(
                    __file__, '../../environments/multi_mode_with_resources/double_adder_mode.yaml'
                ),
                '-p',
                'multi_mode_with_resources',  # pipeline name
            ],
        )
        assert bad_res.exit_code == 2


@pytest.mark.parametrize('gen_execute_args', pipeline_python_origin_contexts())
def test_execute_command_no_env(gen_execute_args):
    with gen_execute_args as (cli_args, instance):
        execute_execute_command(kwargs=cli_args, instance=instance)


@pytest.mark.parametrize('gen_execute_args', pipeline_python_origin_contexts())
def test_execute_command_env(gen_execute_args):
    with gen_execute_args as (cli_args, instance):
        kwargs = merge_dicts(
            {'config': (file_relative_path(__file__, 'default_log_error_env.yaml'),)}, cli_args,
        )
        execute_execute_command(
            kwargs=kwargs, instance=instance,
        )


@pytest.mark.parametrize('cli_args', valid_pipeline_python_origin_target_cli_args())
def test_execute_command_runner(cli_args):
    runner = CliRunner()
    with instance_for_test():
        runner_pipeline_execute(runner, cli_args)

        runner_pipeline_execute(
            runner,
            ['--config', file_relative_path(__file__, 'default_log_error_env.yaml')] + cli_args,
        )


def test_output_execute_log_stdout(capfd):
    with instance_for_test(
        overrides={
            'compute_logs': {
                'module': 'dagster.core.storage.noop_compute_log_manager',
                'class': 'NoOpComputeLogManager',
            }
        },
    ) as instance:
        execute_execute_command(
            kwargs={
                'python_file': file_relative_path(__file__, 'test_cli_commands.py'),
                'attribute': 'stdout_pipeline',
            },
            instance=instance,
        )

        captured = capfd.readouterr()
        # All pipeline execute output currently logged to stderr
        assert 'HELLO WORLD' in captured.err


def test_output_execute_log_stderr(capfd):
    with instance_for_test(
        overrides={
            'compute_logs': {
                'module': 'dagster.core.storage.noop_compute_log_manager',
                'class': 'NoOpComputeLogManager',
            }
        },
    ) as instance:
        with pytest.raises(click.ClickException, match=re.escape('resulted in failure')):
            execute_execute_command(
                kwargs={
                    'python_file': file_relative_path(__file__, 'test_cli_commands.py'),
                    'attribute': 'stderr_pipeline',
                },
                instance=instance,
            )
        captured = capfd.readouterr()
        assert 'I AM SUPPOSED TO FAIL' in captured.err


def test_more_than_one_pipeline():
    with instance_for_test() as instance:
        with pytest.raises(
            UsageError,
            match=re.escape(
                "Must provide --pipeline as there is more than one pipeline in bar. "
                "Options are: ['baz', 'foo']."
            ),
        ):
            execute_execute_command(
                kwargs={
                    'repository_yaml': None,
                    'pipeline': None,
                    'python_file': file_relative_path(__file__, 'test_cli_commands.py'),
                    'module_name': None,
                    'attribute': None,
                },
                instance=instance,
            )


def test_attribute_not_found():
    with instance_for_test() as instance:
        with pytest.raises(
            DagsterInvariantViolationError,
            match=re.escape('nope not found at module scope in file'),
        ):
            execute_execute_command(
                kwargs={
                    'repository_yaml': None,
                    'pipeline': None,
                    'python_file': file_relative_path(__file__, 'test_cli_commands.py'),
                    'module_name': None,
                    'attribute': 'nope',
                },
                instance=instance,
            )


def test_attribute_is_wrong_thing():
    with instance_for_test() as instance:
        with pytest.raises(
            DagsterInvariantViolationError,
            match=re.escape(
                'Loadable attributes must be either a PipelineDefinition or a '
                'RepositoryDefinition. Got 123.'
            ),
        ):
            execute_execute_command(
                kwargs={
                    'repository_yaml': None,
                    'pipeline': None,
                    'python_file': file_relative_path(__file__, 'test_cli_commands.py'),
                    'module_name': None,
                    'attribute': 'not_a_repo_or_pipeline',
                },
                instance=instance,
            )


def test_attribute_fn_returns_wrong_thing():
    with instance_for_test() as instance:
        with pytest.raises(
            DagsterInvariantViolationError,
            match=re.escape(
                "Loadable attributes must be either a PipelineDefinition or a RepositoryDefinition."
            ),
        ):
            execute_execute_command(
                kwargs={
                    'repository_yaml': None,
                    'pipeline': None,
                    'python_file': file_relative_path(__file__, 'test_cli_commands.py'),
                    'module_name': None,
                    'attribute': 'not_a_repo_or_pipeline_fn',
                },
                instance=instance,
            )


def runner_pipeline_execute(runner, cli_args):
    result = runner.invoke(pipeline_execute_command, cli_args)
    if result.exit_code != 0:
        # CliRunner captures stdout so printing it out here
        raise Exception(
            (
                'dagster pipeline execute commands with cli_args {cli_args} '
                'returned exit_code {exit_code} with stdout:\n"{stdout}" and '
                '\nresult as string: "{result}"'
            ).format(
                cli_args=cli_args, exit_code=result.exit_code, stdout=result.stdout, result=result
            )
        )
    return result


def test_default_memory_run_storage():
    with instance_for_test() as instance:
        cli_args = {
            'python_file': file_relative_path(__file__, 'test_cli_commands.py'),
            'atribute': 'bar',
            'pipeline': 'foo',
            'module_name': None,
        }
        result = execute_execute_command(kwargs=cli_args, instance=instance)
        assert result.success


def test_override_with_in_memory_storage():
    with instance_for_test() as instance:
        cli_args = {
            'python_file': file_relative_path(__file__, 'test_cli_commands.py'),
            'atribute': 'bar',
            'pipeline': 'foo',
            'module_name': None,
            'config': (file_relative_path(__file__, 'in_memory_env.yaml'),),
        }
        result = execute_execute_command(kwargs=cli_args, instance=instance,)
        assert result.success


def test_override_with_filesystem_storage():
    with instance_for_test() as instance:
        cli_args = {
            'python_file': file_relative_path(__file__, 'test_cli_commands.py'),
            'atribute': 'bar',
            'pipeline': 'foo',
            'module_name': None,
            'config': (file_relative_path(__file__, 'filesystem_env.yaml'),),
        }
        result = execute_execute_command(kwargs=cli_args, instance=instance,)
        assert result.success


def test_multiproc():
    with instance_for_test():
        runner = CliRunner()
        add_result = runner_pipeline_execute(
            runner,
            [
                '-f',
                file_relative_path(__file__, '../../general_tests/test_repository.py'),
                '-a',
                'dagster_test_repository',
                '--preset',
                'multiproc',
                '-p',
                'multi_mode_with_resources',  # pipeline name
            ],
        )
        assert add_result.exit_code == 0

        assert 'PIPELINE_SUCCESS' in add_result.output


def test_multiproc_invalid():
    # force ephemeral instance by removing out DAGSTER_HOME
    runner = CliRunner(env={'DAGSTER_HOME': None})
    add_result = runner.invoke(
        pipeline_execute_command,
        [
            '-f',
            file_relative_path(__file__, '../../general_tests/test_repository.py'),
            '-a',
            'dagster_test_repository',
            '--preset',
            'multiproc',
            '-p',
            'multi_mode_with_resources',  # pipeline name
        ],
    )
    # which is invalid for multiproc
    assert add_result.exit_code != 0
    assert 'DagsterUnmetExecutorRequirementsError' in add_result.output


def test_tags_pipeline():
    runner = CliRunner()
    with instance_for_test() as instance:
        result = runner.invoke(
            pipeline_execute_command,
            [
                '-m',
                'dagster_tests.cli_tests.command_tests.test_cli_commands',
                '-a',
                'bar',
                '--tags',
                '{ "foo": "bar" }',
                '-p',
                'foo',
            ],
        )
        assert result.exit_code == 0
        runs = instance.get_runs()
        assert len(runs) == 1
        run = runs[0]
        assert len(run.tags) == 1
        assert run.tags.get('foo') == 'bar'

    with instance_for_test() as instance:
        result = runner.invoke(
            pipeline_execute_command,
            [
                '-f',
                file_relative_path(__file__, '../../general_tests/test_repository.py'),
                '-a',
                'dagster_test_repository',
                '--preset',
                'add',
                '--tags',
                '{ "foo": "bar" }',
                '-p',
                'multi_mode_with_resources',  # pipeline name
            ],
        )
        assert result.exit_code == 0
        runs = instance.get_runs()
        assert len(runs) == 1
        run = runs[0]
        assert len(run.tags) == 1
        assert run.tags.get('foo') == 'bar'


def test_execute_subset_pipeline_single_clause_solid_name():
    runner = CliRunner()
    with instance_for_test() as instance:
        result = runner.invoke(
            pipeline_execute_command,
            [
                '-f',
                file_relative_path(__file__, 'test_cli_commands.py'),
                '-a',
                'foo_pipeline',
                '--solid-selection',
                'do_something',
            ],
        )
        assert result.exit_code == 0
        runs = instance.get_runs()
        assert len(runs) == 1
        run = runs[0]
        assert run.solid_selection == ['do_something']
        assert run.solids_to_execute == {'do_something'}


def test_execute_subset_pipeline_single_clause_dsl():
    runner = CliRunner()
    with instance_for_test() as instance:
        result = runner.invoke(
            pipeline_execute_command,
            [
                '-f',
                file_relative_path(__file__, 'test_cli_commands.py'),
                '-a',
                'foo_pipeline',
                '--solid-selection',
                '*do_something+',
            ],
        )
        assert result.exit_code == 0
        runs = instance.get_runs()
        assert len(runs) == 1
        run = runs[0]
        assert run.solid_selection == ['*do_something+']
        assert run.solids_to_execute == {'do_something', 'do_input'}


def test_execute_subset_pipeline_multiple_clauses_dsl_and_solid_name():
    runner = CliRunner()
    with instance_for_test() as instance:
        result = runner.invoke(
            pipeline_execute_command,
            [
                '-f',
                file_relative_path(__file__, 'test_cli_commands.py'),
                '-a',
                'foo_pipeline',
                '--solid-selection',
                '*do_something+,do_input',
            ],
        )
        assert result.exit_code == 0
        runs = instance.get_runs()
        assert len(runs) == 1
        run = runs[0]
        assert set(run.solid_selection) == set(['*do_something+', 'do_input'])
        assert run.solids_to_execute == {'do_something', 'do_input'}


def test_execute_subset_pipeline_invalid():
    runner = CliRunner()
    with instance_for_test():
        result = runner.invoke(
            pipeline_execute_command,
            [
                '-f',
                file_relative_path(__file__, 'test_cli_commands.py'),
                '-a',
                'foo_pipeline',
                '--solid-selection',
                'a, b',
            ],
        )
        assert result.exit_code == 1
        assert 'No qualified solids to execute found for solid_selection' in str(result.exception)



'''
Repository of test pipelines
'''

from dagster import (
    Int,
    ModeDefinition,
    PipelineDefinition,
    PresetDefinition,
    repository,
    resource,
    solid,
)
from dagster.utils import file_relative_path


def define_empty_pipeline():
    return PipelineDefinition(name='empty_pipeline', solid_defs=[])


def define_single_mode_pipeline():
    @solid
    def return_two(_context):
        return 2

    return PipelineDefinition(
        name='single_mode', solid_defs=[return_two], mode_defs=[ModeDefinition(name='the_mode')]
    )


def define_multi_mode_pipeline():
    @solid
    def return_three(_context):
        return 3

    return PipelineDefinition(
        name='multi_mode',
        solid_defs=[return_three],
        mode_defs=[ModeDefinition(name='mode_one'), ModeDefinition('mode_two')],
    )


def define_multi_mode_with_resources_pipeline():
    # API red alert. One has to wrap a type in Field because it is callable
    @resource(config_schema=Int)
    def adder_resource(init_context):
        return lambda x: x + init_context.resource_config

    @resource(config_schema=Int)
    def multer_resource(init_context):
        return lambda x: x * init_context.resource_config

    @resource(config_schema={'num_one': Int, 'num_two': Int})
    def double_adder_resource(init_context):
        return (
            lambda x: x
            + init_context.resource_config['num_one']
            + init_context.resource_config['num_two']
        )

    @solid(required_resource_keys={'op'})
    def apply_to_three(context):
        return context.resources.op(3)

    return PipelineDefinition(
        name='multi_mode_with_resources',
        solid_defs=[apply_to_three],
        mode_defs=[
            ModeDefinition(name='add_mode', resource_defs={'op': adder_resource}),
            ModeDefinition(name='mult_mode', resource_defs={'op': multer_resource}),
            ModeDefinition(
                name='double_adder_mode',
                resource_defs={'op': double_adder_resource},
                description='Mode that adds two numbers to thing',
            ),
        ],
        preset_defs=[
            PresetDefinition.from_files(
                'add',
                mode='add_mode',
                config_files=[
                    file_relative_path(
                        __file__, '../environments/multi_mode_with_resources/add_mode.yaml'
                    )
                ],
            ),
            PresetDefinition(
                'multiproc',
                mode='add_mode',
                run_config={
                    'resources': {'op': {'config': 2}},
                    'execution': {'multiprocess': {}},
                    'storage': {'filesystem': {}},
                },
            ),
        ],
    )


@repository
def dagster_test_repository():
    return [
        define_empty_pipeline(),
        define_single_mode_pipeline(),
        define_multi_mode_pipeline(),
        define_multi_mode_with_resources_pipeline(),
    ]


def test_repository_construction():
    assert dagster_test_repository


@repository
def empty_repository():
    return []



import sys

import pytest

from dagster import seven
from dagster.api.list_repositories import (
    sync_list_repositories,
    sync_list_repositories_ephemeral_grpc,
    sync_list_repositories_grpc,
)
from dagster.core.code_pointer import FileCodePointer, ModuleCodePointer
from dagster.core.errors import DagsterUserCodeProcessError
from dagster.grpc.types import LoadableRepositorySymbol
from dagster.utils import file_relative_path


def test_sync_list_python_file():
    python_file = file_relative_path(__file__, 'api_tests_repo.py')
    loadable_repo_symbols = sync_list_repositories(
        sys.executable,
        python_file=python_file,
        module_name=None,
        working_directory=None,
        attribute=None,
    ).repository_symbols

    assert isinstance(loadable_repo_symbols, list)
    assert len(loadable_repo_symbols) == 1
    assert isinstance(loadable_repo_symbols[0], LoadableRepositorySymbol)

    symbol = loadable_repo_symbols[0]

    assert symbol.repository_name == 'bar_repo'
    assert symbol.attribute == 'bar_repo'


def test_sync_list_python_file_multi_repo():
    python_file = file_relative_path(__file__, 'multiple_repos.py')
    loadable_repo_symbols = sync_list_repositories(
        sys.executable,
        python_file=python_file,
        module_name=None,
        working_directory=None,
        attribute=None,
    ).repository_symbols

    assert isinstance(loadable_repo_symbols, list)
    assert len(loadable_repo_symbols) == 2
    assert isinstance(loadable_repo_symbols[0], LoadableRepositorySymbol)
    assert isinstance(loadable_repo_symbols[1], LoadableRepositorySymbol)

    by_symbol = {lrs.attribute: lrs for lrs in loadable_repo_symbols}

    assert by_symbol['repo_one_symbol'].repository_name == 'repo_one'
    assert by_symbol['repo_two'].repository_name == 'repo_two'


def test_sync_list_python_file_attribute_multi_repo():
    python_file = file_relative_path(__file__, 'multiple_repos.py')
    loadable_repo_symbols = sync_list_repositories(
        sys.executable,
        python_file=python_file,
        module_name=None,
        working_directory=None,
        attribute='repo_one_symbol',
    ).repository_symbols

    assert isinstance(loadable_repo_symbols, list)
    assert len(loadable_repo_symbols) == 1
    assert isinstance(loadable_repo_symbols[0], LoadableRepositorySymbol)

    assert loadable_repo_symbols[0].repository_name == 'repo_one'
    assert loadable_repo_symbols[0].attribute == 'repo_one_symbol'


def test_sync_list_python_module():
    loadable_repo_symbols = sync_list_repositories(
        sys.executable,
        python_file=None,
        module_name='dagster.utils.test.hello_world_repository',
        working_directory=None,
        attribute=None,
    ).repository_symbols

    assert isinstance(loadable_repo_symbols, list)
    assert len(loadable_repo_symbols) == 1
    assert isinstance(loadable_repo_symbols[0], LoadableRepositorySymbol)

    symbol = loadable_repo_symbols[0]

    assert symbol.repository_name == 'hello_world_repository'
    assert symbol.attribute == 'hello_world_repository'


def test_sync_list_python_module_attribute():
    loadable_repo_symbols = sync_list_repositories(
        sys.executable,
        python_file=None,
        module_name='dagster.utils.test.hello_world_repository',
        working_directory=None,
        attribute='hello_world_repository',
    ).repository_symbols

    assert isinstance(loadable_repo_symbols, list)
    assert len(loadable_repo_symbols) == 1
    assert isinstance(loadable_repo_symbols[0], LoadableRepositorySymbol)

    symbol = loadable_repo_symbols[0]

    assert symbol.repository_name == 'hello_world_repository'
    assert symbol.attribute == 'hello_world_repository'


def test_sync_list_python_file_grpc():
    python_file = file_relative_path(__file__, 'api_tests_repo.py')
    response = sync_list_repositories_ephemeral_grpc(
        sys.executable,
        python_file=python_file,
        module_name=None,
        working_directory=None,
        attribute=None,
    )

    loadable_repo_symbols = response.repository_symbols

    assert isinstance(loadable_repo_symbols, list)
    assert len(loadable_repo_symbols) == 1
    assert isinstance(loadable_repo_symbols[0], LoadableRepositorySymbol)

    symbol = loadable_repo_symbols[0]

    assert symbol.repository_name == 'bar_repo'
    assert symbol.attribute == 'bar_repo'

    executable_path = response.executable_path
    assert executable_path == sys.executable

    repository_code_pointer_dict = response.repository_code_pointer_dict
    assert 'bar_repo' in repository_code_pointer_dict
    assert isinstance(repository_code_pointer_dict['bar_repo'], FileCodePointer)
    assert repository_code_pointer_dict['bar_repo'].python_file.endswith("api_tests_repo.py")
    assert repository_code_pointer_dict['bar_repo'].fn_name == 'bar_repo'


def test_sync_list_python_file_multi_repo_grpc():
    python_file = file_relative_path(__file__, 'multiple_repos.py')
    response = sync_list_repositories_ephemeral_grpc(
        sys.executable,
        python_file=python_file,
        module_name=None,
        working_directory=None,
        attribute=None,
    )

    loadable_repo_symbols = response.repository_symbols

    assert isinstance(loadable_repo_symbols, list)
    assert len(loadable_repo_symbols) == 2
    assert isinstance(loadable_repo_symbols[0], LoadableRepositorySymbol)
    assert isinstance(loadable_repo_symbols[1], LoadableRepositorySymbol)

    by_symbol = {lrs.attribute: lrs for lrs in loadable_repo_symbols}

    assert by_symbol['repo_one_symbol'].repository_name == 'repo_one'
    assert by_symbol['repo_two'].repository_name == 'repo_two'

    executable_path = response.executable_path
    assert executable_path == sys.executable

    repository_code_pointer_dict = response.repository_code_pointer_dict
    assert 'repo_one' in repository_code_pointer_dict
    assert 'repo_two' in repository_code_pointer_dict

    assert isinstance(repository_code_pointer_dict['repo_one'], FileCodePointer)
    assert repository_code_pointer_dict['repo_one'].python_file == python_file
    assert repository_code_pointer_dict['repo_one'].fn_name == 'repo_one_symbol'

    assert isinstance(repository_code_pointer_dict['repo_two'], FileCodePointer)
    assert repository_code_pointer_dict['repo_two'].fn_name == 'repo_two'


def test_sync_list_python_file_attribute_multi_repo_grpc():
    python_file = file_relative_path(__file__, 'multiple_repos.py')
    response = sync_list_repositories_ephemeral_grpc(
        sys.executable,
        python_file=python_file,
        module_name=None,
        working_directory=None,
        attribute='repo_one_symbol',
    )

    loadable_repo_symbols = response.repository_symbols

    assert isinstance(loadable_repo_symbols, list)
    assert len(loadable_repo_symbols) == 1
    assert isinstance(loadable_repo_symbols[0], LoadableRepositorySymbol)

    symbol = loadable_repo_symbols[0]

    assert symbol.repository_name == 'repo_one'
    assert symbol.attribute == 'repo_one_symbol'

    executable_path = response.executable_path
    assert executable_path == sys.executable

    repository_code_pointer_dict = response.repository_code_pointer_dict
    assert 'repo_one' in repository_code_pointer_dict

    assert isinstance(repository_code_pointer_dict['repo_one'], FileCodePointer)
    assert repository_code_pointer_dict['repo_one'].python_file == python_file
    assert repository_code_pointer_dict['repo_one'].fn_name == 'repo_one_symbol'


def test_sync_list_python_module_grpc():
    module_name = 'dagster.utils.test.hello_world_repository'
    response = sync_list_repositories_ephemeral_grpc(
        sys.executable,
        python_file=None,
        module_name=module_name,
        working_directory=None,
        attribute=None,
    )

    loadable_repo_symbols = response.repository_symbols

    assert isinstance(loadable_repo_symbols, list)
    assert len(loadable_repo_symbols) == 1
    assert isinstance(loadable_repo_symbols[0], LoadableRepositorySymbol)

    symbol = loadable_repo_symbols[0]

    assert symbol.repository_name == 'hello_world_repository'
    assert symbol.attribute == 'hello_world_repository'

    executable_path = response.executable_path
    assert executable_path == sys.executable

    repository_code_pointer_dict = response.repository_code_pointer_dict

    assert 'hello_world_repository' in repository_code_pointer_dict
    assert isinstance(repository_code_pointer_dict['hello_world_repository'], ModuleCodePointer)
    assert repository_code_pointer_dict['hello_world_repository'].module == module_name
    assert (
        repository_code_pointer_dict['hello_world_repository'].fn_name == 'hello_world_repository'
    )


def test_sync_list_python_module_attribute_grpc():
    module_name = 'dagster.utils.test.hello_world_repository'
    response = sync_list_repositories_ephemeral_grpc(
        sys.executable,
        python_file=None,
        module_name=module_name,
        working_directory=None,
        attribute='hello_world_repository',
    )

    loadable_repo_symbols = response.repository_symbols

    assert isinstance(loadable_repo_symbols, list)
    assert len(loadable_repo_symbols) == 1
    assert isinstance(loadable_repo_symbols[0], LoadableRepositorySymbol)

    symbol = loadable_repo_symbols[0]

    assert symbol.repository_name == 'hello_world_repository'
    assert symbol.attribute == 'hello_world_repository'

    executable_path = response.executable_path
    assert executable_path == sys.executable

    repository_code_pointer_dict = response.repository_code_pointer_dict

    assert 'hello_world_repository' in repository_code_pointer_dict
    assert isinstance(repository_code_pointer_dict['hello_world_repository'], ModuleCodePointer)
    assert repository_code_pointer_dict['hello_world_repository'].module == module_name
    assert (
        repository_code_pointer_dict['hello_world_repository'].fn_name == 'hello_world_repository'
    )


@pytest.mark.skipif(seven.IS_WINDOWS, reason='Depends on Docker, so skip running in Windows')
def test_sync_list_container_grpc(docker_grpc_client):
    response = sync_list_repositories_grpc(docker_grpc_client)

    loadable_repo_symbols = response.repository_symbols

    assert isinstance(loadable_repo_symbols, list)
    assert len(loadable_repo_symbols) == 1
    assert isinstance(loadable_repo_symbols[0], LoadableRepositorySymbol)

    symbol = loadable_repo_symbols[0]

    assert symbol.repository_name == 'bar_repo'
    assert symbol.attribute == 'bar_repo'

    executable_path = response.executable_path
    assert executable_path

    repository_code_pointer_dict = response.repository_code_pointer_dict
    assert 'bar_repo' in repository_code_pointer_dict
    assert isinstance(repository_code_pointer_dict['bar_repo'], FileCodePointer)
    assert repository_code_pointer_dict['bar_repo'].python_file.endswith("repo.py")
    assert repository_code_pointer_dict['bar_repo'].fn_name == 'bar_repo'


def test_sync_list_python_file_with_error():
    python_file = file_relative_path(__file__, 'error_on_load_repo.py')
    with pytest.raises(DagsterUserCodeProcessError) as e:
        sync_list_repositories(
            sys.executable,
            python_file=python_file,
            module_name=None,
            working_directory=None,
            attribute=None,
        )

    assert e.value.args[0].startswith('(ValueError) - ValueError: User did something bad')
    assert e.value.args[0].endswith("raise ValueError('User did something bad')\n")


def test_sync_list_python_file_grpc_with_error():
    python_file = file_relative_path(__file__, 'error_on_load_repo.py')
    with pytest.raises(DagsterUserCodeProcessError) as e:
        sync_list_repositories_ephemeral_grpc(
            sys.executable,
            python_file=python_file,
            module_name=None,
            working_directory=None,
            attribute=None,
        )

    assert e.value.args[0].startswith('(ValueError) - ValueError: User did something bad')
    assert e.value.args[0].endswith("raise ValueError('User did something bad')\n")



import os
import time

import pytest

from dagster.core.executor.child_process_executor import (
    ChildProcessCommand,
    ChildProcessCrashException,
    ChildProcessDoneEvent,
    ChildProcessEvent,
    ChildProcessStartEvent,
    ChildProcessSystemErrorEvent,
    execute_child_process_command,
)
from dagster.utils import segfault


class DoubleAStringChildProcessCommand(ChildProcessCommand):
    def __init__(self, a_str):
        self.a_str = a_str

    def execute(self):
        yield self.a_str + self.a_str


class AnError(Exception):
    pass


class ThrowAnErrorCommand(ChildProcessCommand):  # pylint: disable=no-init
    def execute(self):
        raise AnError('Oh noes!')


class CrashyCommand(ChildProcessCommand):  # pylint: disable=no-init
    def execute(self):
        # access inner API to simulate hard crash
        os._exit(1)  # pylint: disable=protected-access


class SegfaultCommand(ChildProcessCommand):  # pylint: disable=no-init
    def execute(self):
        # access inner API to simulate hard crash
        segfault()


class LongRunningCommand(ChildProcessCommand):  # pylint: disable=no-init
    def execute(self):
        time.sleep(0.5)
        yield 1


def test_basic_child_process_command():
    events = list(
        filter(
            lambda x: x and not isinstance(x, ChildProcessEvent),
            execute_child_process_command(DoubleAStringChildProcessCommand('aa')),
        )
    )
    assert events == ['aaaa']


def test_basic_child_process_command_with_process_events():
    events = list(
        filter(lambda x: x, execute_child_process_command(DoubleAStringChildProcessCommand('aa')))
    )
    assert len(events) == 3

    assert isinstance(events[0], ChildProcessStartEvent)
    child_pid = events[0].pid
    assert child_pid != os.getpid()
    assert events[1] == 'aaaa'
    assert isinstance(events[2], ChildProcessDoneEvent)
    assert events[2].pid == child_pid


def test_child_process_uncaught_exception():
    results = list(
        filter(
            lambda x: x and isinstance(x, ChildProcessSystemErrorEvent),
            execute_child_process_command(ThrowAnErrorCommand()),
        )
    )
    assert len(results) == 1

    assert 'AnError' in str(results[0].error_info.message)


def test_child_process_crashy_process():
    with pytest.raises(ChildProcessCrashException) as exc:
        list(execute_child_process_command(CrashyCommand()))
    assert exc.value.exit_code == 1


@pytest.mark.skipif(os.name == 'nt', reason="Segfault not being caught on Windows: See issue #2791")
def test_child_process_segfault():
    with pytest.raises(ChildProcessCrashException) as exc:
        list(execute_child_process_command(SegfaultCommand()))
    assert exc.value.exit_code == -11


@pytest.mark.skip('too long')
def test_long_running_command():
    list(execute_child_process_command(LongRunningCommand()))



from __future__ import print_function

import re

import pytest
from click.testing import CliRunner

from dagster.cli.pipeline import execute_print_command, pipeline_print_command
from dagster.utils import file_relative_path

from .test_cli_commands import (
    launch_command_contexts,
    valid_external_pipeline_target_cli_args_no_preset,
)


def no_print(_):
    return None


@pytest.mark.parametrize('gen_pipeline_args', launch_command_contexts())
def test_print_command_verbose(gen_pipeline_args):
    with gen_pipeline_args as (cli_args, uses_legacy_repository_yaml_format, instance):
        if uses_legacy_repository_yaml_format:
            with pytest.warns(
                UserWarning,
                match=re.escape(
                    'You are using the legacy repository yaml format. Please update your file '
                ),
            ):
                execute_print_command(
                    verbose=True, cli_args=cli_args, print_fn=no_print, instance=instance
                )
        else:
            execute_print_command(
                verbose=True, cli_args=cli_args, print_fn=no_print, instance=instance
            )


@pytest.mark.parametrize('gen_pipeline_args', launch_command_contexts())
def test_print_command(gen_pipeline_args):
    with gen_pipeline_args as (cli_args, uses_legacy_repository_yaml_format, instance):
        if uses_legacy_repository_yaml_format:
            with pytest.warns(
                UserWarning,
                match=re.escape(
                    'You are using the legacy repository yaml format. Please update your file '
                ),
            ):
                execute_print_command(
                    verbose=False, cli_args=cli_args, print_fn=no_print, instance=instance
                )
        else:
            execute_print_command(
                verbose=False, cli_args=cli_args, print_fn=no_print, instance=instance
            )


@pytest.mark.parametrize('pipeline_cli_args', valid_external_pipeline_target_cli_args_no_preset())
def test_print_command_cli(pipeline_cli_args):
    runner = CliRunner()
    cli_args, uses_legacy_repository_yaml_format = pipeline_cli_args
    if uses_legacy_repository_yaml_format:
        with pytest.warns(
            UserWarning,
            match=re.escape(
                'You are using the legacy repository yaml format. Please update your file '
            ),
        ):
            result = runner.invoke(pipeline_print_command, cli_args)
            assert result.exit_code == 0, result.stdout

            result = runner.invoke(pipeline_print_command, ['--verbose'] + cli_args)
            assert result.exit_code == 0, result.stdout
    else:
        result = runner.invoke(pipeline_print_command, cli_args)
        assert result.exit_code == 0, result.stdout

        result = runner.invoke(pipeline_print_command, ['--verbose'] + cli_args)
        assert result.exit_code == 0, result.stdout


def test_print_command_baz():
    runner = CliRunner()
    res = runner.invoke(
        pipeline_print_command,
        [
            '--verbose',
            '-f',
            file_relative_path(__file__, 'test_cli_commands.py'),
            '-a',
            'bar',
            '-p',
            'baz',
        ],
    )
    assert res.exit_code == 0, res.stdout



from __future__ import print_function

import mock
import pytest
from click.testing import CliRunner

from dagster.cli.schedule import (
    schedule_list_command,
    schedule_logs_command,
    schedule_restart_command,
    schedule_start_command,
    schedule_stop_command,
    schedule_up_command,
    schedule_wipe_command,
)

from .test_cli_commands import schedule_command_contexts


@pytest.mark.parametrize('gen_schedule_args', schedule_command_contexts())
def test_schedules_list(gen_schedule_args):
    with gen_schedule_args as (cli_args, instance):
        runner = CliRunner()
        with mock.patch('dagster.core.instance.DagsterInstance.get') as _instance:
            _instance.return_value = instance
            result = runner.invoke(schedule_list_command, cli_args)

            if result.exception:
                raise result.exception

            assert result.exit_code == 0
            assert result.output == ('Repository bar\n' '**************\n')


@pytest.mark.parametrize('gen_schedule_args', schedule_command_contexts())
def test_schedules_up(gen_schedule_args):
    with gen_schedule_args as (cli_args, instance):
        runner = CliRunner()
        with mock.patch('dagster.core.instance.DagsterInstance.get') as _instance:
            _instance.return_value = instance
            result = runner.invoke(schedule_up_command, cli_args,)
            assert result.exit_code == 0
            assert 'Changes:\n  + foo_schedule (add)' in result.output


@pytest.mark.parametrize('gen_schedule_args', schedule_command_contexts())
def test_schedules_up_and_list(gen_schedule_args):
    with gen_schedule_args as (cli_args, instance):
        runner = CliRunner()
        with mock.patch('dagster.core.instance.DagsterInstance.get') as _instance:
            _instance.return_value = instance

            result = runner.invoke(schedule_up_command, cli_args)

            result = runner.invoke(schedule_list_command, cli_args)

            assert result.exit_code == 0
            assert (
                result.output == 'Repository bar\n'
                '**************\n'
                'Schedule: foo_schedule [STOPPED]\n'
                'Cron Schedule: * * * * *\n'
            )


@pytest.mark.parametrize('gen_schedule_args', schedule_command_contexts())
def test_schedules_start_and_stop(gen_schedule_args):
    with gen_schedule_args as (cli_args, instance):
        with mock.patch('dagster.core.instance.DagsterInstance.get') as _instance:
            _instance.return_value = instance
            runner = CliRunner()

            result = runner.invoke(schedule_up_command, cli_args,)

            result = runner.invoke(schedule_start_command, cli_args + ['foo_schedule'],)

            assert result.exit_code == 0
            assert 'Started schedule foo_schedule\n' == result.output

            result = runner.invoke(schedule_stop_command, cli_args + ['foo_schedule'],)

            assert result.exit_code == 0
            assert 'Stopped schedule foo_schedule\n' == result.output


@pytest.mark.parametrize('gen_schedule_args', schedule_command_contexts())
def test_schedules_start_empty(gen_schedule_args):
    with gen_schedule_args as (cli_args, instance):
        runner = CliRunner()
        with mock.patch('dagster.core.instance.DagsterInstance.get') as _instance:
            _instance.return_value = instance
            result = runner.invoke(schedule_start_command, cli_args,)

            assert result.exit_code == 0
            assert 'Noop: dagster schedule start was called without any arguments' in result.output


@pytest.mark.parametrize('gen_schedule_args', schedule_command_contexts())
def test_schedules_start_all(gen_schedule_args):
    with gen_schedule_args as (cli_args, instance):
        runner = CliRunner()
        with mock.patch('dagster.core.instance.DagsterInstance.get') as _instance:
            _instance.return_value = instance
            result = runner.invoke(schedule_up_command, cli_args)

            result = runner.invoke(schedule_start_command, cli_args + ['--start-all'],)

            assert result.exit_code == 0
            assert result.output == 'Started all schedules for repository bar\n'


@pytest.mark.parametrize('gen_schedule_args', schedule_command_contexts())
def test_schedules_wipe_correct_delete_message(gen_schedule_args):
    with gen_schedule_args as (cli_args, instance):
        runner = CliRunner()
        with mock.patch('dagster.core.instance.DagsterInstance.get') as _instance:
            _instance.return_value = instance
            result = runner.invoke(schedule_up_command, cli_args)

            result = runner.invoke(schedule_wipe_command, cli_args, input="DELETE\n",)

            if result.exception:
                raise result.exception

            assert result.exit_code == 0
            assert 'Wiped all schedules and schedule cron jobs' in result.output

            result = runner.invoke(schedule_up_command, cli_args + ['--preview'],)

            # Verify schedules were wiped
            assert result.exit_code == 0
            assert 'Planned Schedule Changes:\n  + foo_schedule (add)' in result.output


@pytest.mark.parametrize('gen_schedule_args', schedule_command_contexts())
def test_schedules_wipe_incorrect_delete_message(gen_schedule_args):
    with gen_schedule_args as (cli_args, instance):
        runner = CliRunner()
        with mock.patch('dagster.core.instance.DagsterInstance.get') as _instance:
            _instance.return_value = instance
            result = runner.invoke(schedule_up_command, cli_args)

            result = runner.invoke(schedule_wipe_command, cli_args, input="WRONG\n",)

            assert result.exit_code == 0
            assert 'Exiting without deleting all schedules and schedule cron jobs' in result.output

            result = runner.invoke(schedule_up_command, cli_args + ['--preview'],)

            # Verify schedules were not wiped
            assert result.exit_code == 0
            assert (
                result.output
                == 'No planned changes to schedules.\n1 schedules will remain unchanged\n'
            )


@pytest.mark.parametrize('gen_schedule_args', schedule_command_contexts())
def test_schedules_restart(gen_schedule_args):
    with gen_schedule_args as (cli_args, instance):
        runner = CliRunner()
        with mock.patch('dagster.core.instance.DagsterInstance.get') as _instance:
            _instance.return_value = instance

            result = runner.invoke(schedule_up_command, cli_args)

            result = runner.invoke(schedule_start_command, cli_args + ['foo_schedule'],)

            result = runner.invoke(schedule_restart_command, cli_args + ['foo_schedule'],)

            assert result.exit_code == 0
            assert 'Restarted schedule foo_schedule' in result.output


@pytest.mark.parametrize('gen_schedule_args', schedule_command_contexts())
def test_schedules_restart_all(gen_schedule_args):
    with gen_schedule_args as (cli_args, instance):
        runner = CliRunner()
        with mock.patch('dagster.core.instance.DagsterInstance.get') as _instance:
            _instance.return_value = instance

            result = runner.invoke(schedule_up_command, cli_args)

            result = runner.invoke(schedule_start_command, cli_args + ['foo_schedule'],)

            result = runner.invoke(
                schedule_restart_command, cli_args + ['foo_schedule', '--restart-all-running'],
            )
            assert result.exit_code == 0
            assert result.output == 'Restarted all running schedules for repository bar\n'


@pytest.mark.parametrize('gen_schedule_args', schedule_command_contexts())
def test_schedules_logs(gen_schedule_args):
    with gen_schedule_args as (cli_args, instance):
        with mock.patch('dagster.core.instance.DagsterInstance.get') as _instance:
            _instance.return_value = instance
            runner = CliRunner()

            result = runner.invoke(schedule_logs_command, cli_args + ['foo_schedule'],)

            assert result.exit_code == 0
            assert result.output.endswith('scheduler.log\n')
